import boto3
import datetime
import time
from botocore.exceptions import ClientError
from . import ec2, elb, elb2, cw, autoscaling

PREF = {
	'CPUUtilization': ('Average', 'Percent'),
	'CPUCreditBalance': ('Average', 'Count'),
	'CPUCreditUsage': ('Average', 'Count'),
	'DiskReadBytes': ('Average', 'Bytes'),
	'DiskReadOps': ('Sum', 'Bytes'),
	'DiskWriteBytes': ('Average', 'Bytes'),
	'DiskWriteOps': ('Sum', 'Bytes'),
	'NetworkIn': ('Maximum', 'Bytes'),
	'NetworkOut': ('Maximum', 'Bytes'),
}

class ELBOps:
	def __init__ (self, system, image_id, region = "us-east-1"):
		self.system = system
		self.image_id = image_id
		self.region = region

	def get_metric (self, instance_id, metric = 'CPUUtilization'):
		Statistics, Unit = PREF [metric]
		s = cw.get_metric_statistics(
			Period = 300,
			StartTime = datetime.datetime.utcnow() - datetime.timedelta(seconds=1200),
			EndTime = datetime.datetime.utcnow(),
			MetricName = metric,
			Namespace = 'AWS/EC2',
			Statistics = [Statistics],
			Unit = Unit,
			Dimensions = [{'Name':'InstanceId', 'Value': instance_id}]
		)
		return [p [Statistics] for p in s ["Datapoints"][:3]]

	def create_instance (self, itype, user_data, key_name):
		instances = ec2.create_instances (
			ImageId=self.image_id,
			MinCount = 1,
			MaxCount = 1,
			InstanceType = itype,
			Monitoring = {'Enabled': False},
			SecurityGroups = ['default', '%s-app' % self.system],
			UserData = user_data,
			DryRun=False,
			KeyName = key_name,
			IamInstanceProfile = {'Name': 'app'},
			InstanceInitiatedShutdownBehavior = 'terminate',
			TagSpecifications = [
				{
					'ResourceType': 'instance',
					'Tags': [
							{'Key': 'Name', 'Value': '%s-app-%s' % (self.system, self.image_id [4:])},
							{'Key': 'system', 'Value': self.system},
					]
				},
		  ]
		)
		instance = instances [0]
		instance.wait_until_running ()
		return instance.id

	def stop (self, *ids):
		self.elb_exists () and self.deregister_instances (*ids)
		ec2.instances.filter(InstanceIds=ids).stop()

	def terminate (self, *ids):
		self.elb_exists () and self.deregister_instances (*ids)
		ec2.instances.filter(InstanceIds=ids).terminate()

	def find_instances_by_image (self, image_id = None):
		return ec2.instances.filter(
			Filters=[{'Name': 'tag:Name', 'Values': ['%s-app-%s' % (
				self.system, (image_id or self.image_id) [4:]
			)]}]
		)

	def find_instances (self):
		return ec2.instances.filter(Filters=[{'Name': 'tag:system', 'Values': [self.system]}])

	def describe_instance_status (self):
		for status in ec2.meta.client.describe_instance_status()['InstanceStatuses']:
			print(status)

	def register_instances2 (self, arn, port = 80, *ids):
		elb2.register_targets (
			TargetGroupArn = arn,
			Targets = [{'Id': i, 'Port': port} for i in ids]
		)

	def deregister_instances2 (self, arn, port = 80, *ids):
		elb2.deregister_targets (
			TargetGroupArn = arn,
			Targets = [{'Id': i, 'Port': port} for i in ids]
		)

	def register_instances (self, *ids):
		elb.register_instances_with_load_balancer(
			LoadBalancerName='%s-app' % self.system,
			Instances=[{'InstanceId': i} for i in ids]
		)

	def deregister_instances (self, *ids):
		elb.deregister_instances_from_load_balancer(
			LoadBalancerName='%s-app' % self.system,
			Instances=[{'InstanceId': i} for i in ids]
		)

	def create_elb (self, security_groups):
		elb.create_load_balancer (
			LoadBalancerName = '%s-app' % self.system,
			AvailabilityZones = [self.region + 'a', self.region + 'b', self.region + 'c', self.region + 'd', self.region + 'e', self.region + 'f'],
			SecurityGroups = security_groups,
			Tags=[{'Key': 'system', 'Value': self.system }],
			Listeners=[
				{
					'Protocol': 'HTTP',
					'LoadBalancerPort': 80,
					'InstanceProtocol': 'HTTP',
					'InstancePort': 80
				},
				{
					'Protocol': 'HTTP',
					'LoadBalancerPort': 5000,
					'InstanceProtocol': 'HTTP',
					'InstancePort': 5000
				}
    		]
		)

	def delete_elb (self):
		elb.delete_load_balancer (
			LoadBalancerName = '%s-app' % self.system
		)

	def elb_exists (self):
		try:
			self.elb_health ()
		except ClientError as e:
			if e.response['Error']['Code'] == "LoadBalancerNotFound":
				return 0
			raise
		return 1

	def elb_health (self):
		h =  elb.describe_instance_health (LoadBalancerName = '%s-app' % self.system)
		return dict ([(i ['InstanceId'], i ['State']) for i in h ["InstanceStates"]])

	def create (self, instance_type = 't2.nano', user_data = None):
		instance_id = self.create_instance (instance_type, user_data)
		self.register_instances (instance_id)



import urllib.request, urllib.parse, urllib.error
import re
import time
from ...client.http import ClientCookie
from ...lib import timeoutsocket, confparse
from ...lib import logger as loggerfactory
import tempfile
import os
import random
from ...lib import pathtool
from . import __init__
import math

n, p = pathtool.modpath (__init__)
DEFAULT_PROXYSITES = os.path.join (os.path.split (p) [0], "proxysite.dat")

rx_ip = r"[^0-9]([0-9]{2,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})"
rx_dns = r"[^a-z0-9]([0-9a-z][-a-z0-9_]{2,30}\.[a-z0-9_]{2,30}\.[a-z0-9_.]*?)"
rx_port = r"([0-9]{2,5})"

RX_PROXY = [
		re.compile ("%s:%s" % (rx_ip, rx_port), re.I),
		re.compile ("%s.+?:%s" % (rx_ip, rx_port)),
		re.compile ("%s</td>\s*?<td.*?>%s</td>" % (rx_ip, rx_port), re.M),	
		re.compile ("%s\s+port\s+%s</" % (rx_ip, rx_port)),	
		re.compile ("%s:%s" % (rx_dns, rx_port), re.I),
		re.compile ("%s</td>\s*?<td.*?>%s</td>" % (rx_dns, rx_port), re.I|re.M),
		re.compile ("%s\s+port\s+%s</" % (rx_dns, rx_port)),		
		re.compile ("%s.+?>\s*%s" % (rx_ip, rx_port), re.I|re.S),
]

def unhide_text (data):
	t=''
	s = urllib.parse.unquote (data)
	x = int (round(math.sqrt(49)))
	for c in s: 		
 		t += chr (ord(c)^x)
 	t = t.replace ("#", " ").replace ("!", " ").replace ("|", " ").replace ("-", "")
 	return t

def unhide_text2 (data):
	t=''
	s = urllib.parse.unquote (data)
	for c in s: 		
 		t += chr (ord(c)^4) 	
 	return t

class ProxyCollector:
	def __init__ (self, logger = None):
		self.logger = logger
		self.proxies = {}
		self.cookies = ClientCookie.CookieJar()
		self.opener = ClientCookie.build_opener(ClientCookie.HTTPCookieProcessor(self.cookies))
		
	def save (self, output = None):
		if output is None:
			output = os.path.join (os.environ ["systemroot"], "sharedproxy.dat")
		proxies = list(self.proxies.keys ())
		f = open (output, "w")		
		for proxy in proxies:
			f.write ("%s:%s\n" % proxy)
		f.close ()
		if self.logger: self.logger ("[info] %d proxies saved" % len (self.proxies))
			
	def find (self, data):
		for rx in RX_PROXY:
			proxies = rx.findall (data)
			count = 0
			for server, port in proxies:
				try: port = int (port)
				except: continue				
				count += 1
				self.proxies [(server, port)] = 0
			if count: break
			
		if self.logger: self.logger ("[info] %d proxies found" % len (self.proxies))
		return count
	
	def find_various (self, data):
		self.find (data)
		self.find (urllib.parse.unquote (data))
		self.find (unhide_text2 (data))
		self.find (unhide_text (data))		
	
	rx_evalcode = re.compile ("::(.+?)::")	
	def collect (self, src = None):
		if src is None:
			src = DEFAULT_PROXYSITES
			
		timeoutsocket.setDefaultSocketTimeout (60)
		cf = confparse.ConfParse (src)
		
		proxypages = cf.getopt ("urls")
		random.shuffle (proxypages)
		for page in cf.getopt ("verify") + proxypages:
			for code in self.rx_evalcode.findall (page):
				res = eval (code)
				page = page.replace ("::"+code+"::", res)
				
			try:
				if self.logger: self.logger ("[info] retrieve %s" % page)
				res = self.opener.open(page)
				data = res.read ()				
				self.find_various (data)
			except:
				if self.logger: 
					self.logger.trace ("proxyfinder")
				else:
					raise	
			else:
				time.sleep (3.0)

		
if __name__ == "__main__":
	f = ProxyCollector ()
	f.collect ()
	f.save ()
	



hangul_full = r'[ᄀ-ᇹㄱ-ㆎ가-힣]'
hiragana_full = r'[ぁ-ゟ]'
katakana_full = r'[゠-ヿ]'
kanji = r'[㐀-䶵一-鿋豈-頻]'
radicals = r'[⺀-⿕]'
katakana_half_width = r'[｟-ﾟ]'
alphanum_full = r'[！-～]'
symbols_punct = r'[、-〿]'
misc_symbols = r'[ㇰ-ㇿ㈠-㉃㊀-㋾㌀-㍿]'
ascii_char = r'[ -~]'



import os
import sys
from rs4.psutil.piped import Piped

class PreciseEngine (Piped):
    def __init__(self, exe_file, model_file, chunk_size=2048):
        self.chunk_size = chunk_size
        cmd = [self.abspath (exe_file)]
        cmd += [self.abspath (model_file), str(self.chunk_size)]
        Piped.__init__ (self, cmd)

    def abspath (self, path):
        return os.path.join (os.path.dirname (os.path.abspath (sys.argv [0])), path)

    def communicate (self, input):
        if len(input) != self.chunk_size:
            raise ValueError('Invalid chunk size: ' + str(len(input)))
        out = Piped.communicate (self, input)
        return float(out)

class TriggerDetector:
    def __init__(self, chunk_size, sensitivity=0.5, trigger_level=3):
        self.chunk_size = chunk_size
        self.configure (sensitivity, trigger_level)
        self.activation = 0

    def reset (self):
        self.activation = 0

    def configure (self, sensitivity, trigger_level):
        self.sensitivity = sensitivity
        self.trigger_level = int (trigger_level * (2048 // self.chunk_size))

    def update(self, prob):
        chunk_activated = prob > 1.0 - self.sensitivity
        if chunk_activated or self.activation < 0:
            self.activation += 1
            has_activated = self.activation >= self.trigger_level
            if has_activated or chunk_activated and self.activation < 0:
                self.activation = -(8 * 2048) // self.chunk_size
            if has_activated:
                return True
        elif self.activation > 0:
            self.activation -= 1
        return False

class PreciseRunner(object):
    def __init__(self, model, chunk_size, sensitivity=0.5, trigger_level=4, on_prediction=lambda x: None, engine = None):
        self.chunk_size = chunk_size
        self.trigger_level = trigger_level
        self.on_prediction = on_prediction
        self.engine = self.create_engine (engine, model, chunk_size)

        self.running = False
        self.detector = TriggerDetector(self.chunk_size, sensitivity, trigger_level)

    def configure (self, sensitivity = 0.5, trigger_level = 4):
        self.detector.configure (sensitivity, trigger_level)

    def create_engine (self, engine, model, chunk_size):
        return PreciseEngine (engine or os.path.join (os.path.dirname (__file__), 'engine.py'), model, chunk_size)

    def reset (self):
        self.detector.reset ()

    def start(self):
        self.engine.start()
        self.running = True

    def stop(self):
        self.engine.stop()
        self.running = False

    def detect (self, chunk):
        if not self.running:
            raise SystemError ('engine already stopped')
        prob = self.engine.communicate (chunk)
        self.on_prediction(prob)
        if self.detector.update(prob):
            return True
        return False



#!/usr/bin/env python3
import houndify
import sys



CLIENT_ID = sys.argv[1]
CLIENT_KEY = sys.argv[2]
QUERY = sys.argv[3]

requestInfo = {
  ## Pretend we're at SoundHound HQ.  Set other fields as appropriate
  'Latitude': 37.388309, 
  'Longitude': -121.973968
}

client = houndify.TextHoundClient(CLIENT_ID, CLIENT_KEY, "test_user", requestInfo)

## Uncomment the lines below to see an example of using a custom
## grammar for matching.  Use the file 'turnthelightson.wav' to try it.
# clientMatches = [ {
#   "Expression" : '([1/100 ("can"|"could"|"will"|"would")."you"].[1/10 "please"].("turn"|"switch"|(1/100 "flip"))."on".["the"].("light"|"lights").[1/20 "for"."me"].[1/20 "please"])|([1/100 ("can"|"could"|"will"|"would")."you"].[1/10 "please"].[100 ("turn"|"switch"|(1/100 "flip"))].["the"].("light"|"lights")."on".[1/20 "for"."me"].[1/20 "please"])|((("i".("want"|"like"))|((("i".["would"])|("i\'d")).("like"|"want"))).["the"].("light"|"lights").["turned"|"switched"|("to"."go")|(1/100"flipped")]."on".[1/20"please"])"',
#   "Result" : { "Intent" : "TURN_LIGHT_ON" },
#   "SpokenResponse" : "Ok, I\'m turning the lights on.",
#   "SpokenResponseLong" : "Ok, I\'m turning the lights on.",
#   "WrittenResponse" : "Ok, I\'m turning the lights on.",
#   "WrittenResponseLong" : "Ok, I\'m turning the lights on."
# } ]
# client.setHoundRequestInfo('ClientMatches', clientMatches)

response = client.query(QUERY)
print(response)


# Copyright (C) 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sample that implements a gRPC client for the Google Assistant API."""

import concurrent.futures
import json
import logging
import os
import os.path
import pathlib2 as pathlib
import sys
import time
import uuid

import click
import grpc
import google.auth.transport.grpc
import google.auth.transport.requests
import google.oauth2.credentials

from google.assistant.embedded.v1alpha2 import (
    embedded_assistant_pb2,
    embedded_assistant_pb2_grpc
)
from tenacity import retry, stop_after_attempt, retry_if_exception

try:
    from . import (
        assistant_helpers,
        audio_helpers,
        browser_helpers,
        device_helpers
    )
except (SystemError, ImportError):
    import assistant_helpers
    import audio_helpers
    import browser_helpers
    import device_helpers


ASSISTANT_API_ENDPOINT = 'embeddedassistant.googleapis.com'
END_OF_UTTERANCE = embedded_assistant_pb2.AssistResponse.END_OF_UTTERANCE
DIALOG_FOLLOW_ON = embedded_assistant_pb2.DialogStateOut.DIALOG_FOLLOW_ON
CLOSE_MICROPHONE = embedded_assistant_pb2.DialogStateOut.CLOSE_MICROPHONE
PLAYING = embedded_assistant_pb2.ScreenOutConfig.PLAYING
DEFAULT_GRPC_DEADLINE = 60 * 3 + 5


class SampleAssistant(object):
    """Sample Assistant that supports conversations and device actions.

    Args:
      device_model_id: identifier of the device model.
      device_id: identifier of the registered device instance.
      conversation_stream(ConversationStream): audio stream
        for recording query and playing back assistant answer.
      channel: authorized gRPC channel for connection to the
        Google Assistant API.
      deadline_sec: gRPC deadline in seconds for Google Assistant API call.
      device_handler: callback for device actions.
    """

    def __init__(self, language_code, device_model_id, device_id,
                 conversation_stream, display,
                 channel, deadline_sec, device_handler):
        self.language_code = language_code
        self.device_model_id = device_model_id
        self.device_id = device_id
        self.conversation_stream = conversation_stream
        self.display = display

        # Opaque blob provided in AssistResponse that,
        # when provided in a follow-up AssistRequest,
        # gives the Assistant a context marker within the current state
        # of the multi-Assist()-RPC "conversation".
        # This value, along with MicrophoneMode, supports a more natural
        # "conversation" with the Assistant.
        self.conversation_state = None
        # Force reset of first conversation.
        self.is_new_conversation = True

        # Create Google Assistant API gRPC client.
        self.assistant = embedded_assistant_pb2_grpc.EmbeddedAssistantStub(
            channel
        )
        self.deadline = deadline_sec

        self.device_handler = device_handler

    def __enter__(self):
        return self

    def __exit__(self, etype, e, traceback):
        if e:
            return False
        self.conversation_stream.close()

    def is_grpc_error_unavailable(e):
        is_grpc_error = isinstance(e, grpc.RpcError)
        if is_grpc_error and (e.code() == grpc.StatusCode.UNAVAILABLE):
            logging.error('grpc unavailable error: %s', e)
            return True
        return False

    @retry(reraise=True, stop=stop_after_attempt(3),
           retry=retry_if_exception(is_grpc_error_unavailable))
    def assist(self):
        """Send a voice request to the Assistant and playback the response.

        Returns: True if conversation should continue.
        """
        continue_conversation = False
        device_actions_futures = []

        self.conversation_stream.start_recording()
        logging.info('Recording audio request.')

        def iter_log_assist_requests():
            for c in self.gen_assist_requests():
                assistant_helpers.log_assist_request_without_audio(c)
                yield c
            logging.debug('Reached end of AssistRequest iteration.')

        # This generator yields AssistResponse proto messages
        # received from the gRPC Google Assistant API.
        for resp in self.assistant.Assist(iter_log_assist_requests(),
                                          self.deadline):
            assistant_helpers.log_assist_response_without_audio(resp)
            if resp.event_type == END_OF_UTTERANCE:
                logging.info('End of audio request detected.')
                logging.info('Stopping recording.')
                self.conversation_stream.stop_recording()
            if resp.speech_results:
                logging.info('Transcript of user request: "%s".',
                             ' '.join(r.transcript
                                      for r in resp.speech_results))
            if len(resp.audio_out.audio_data) > 0:
                if not self.conversation_stream.playing:
                    self.conversation_stream.stop_recording()
                    self.conversation_stream.start_playback()
                    logging.info('Playing assistant response.')
                self.conversation_stream.write(resp.audio_out.audio_data)
            if resp.dialog_state_out.conversation_state:
                conversation_state = resp.dialog_state_out.conversation_state
                logging.debug('Updating conversation state.')
                self.conversation_state = conversation_state
            if resp.dialog_state_out.volume_percentage != 0:
                volume_percentage = resp.dialog_state_out.volume_percentage
                logging.info('Setting volume to %s%%', volume_percentage)
                self.conversation_stream.volume_percentage = volume_percentage
            if resp.dialog_state_out.microphone_mode == DIALOG_FOLLOW_ON:
                continue_conversation = True
                logging.info('Expecting follow-on query from user.')
            elif resp.dialog_state_out.microphone_mode == CLOSE_MICROPHONE:
                continue_conversation = False
            if resp.device_action.device_request_json:
                device_request = json.loads(
                    resp.device_action.device_request_json
                )
                fs = self.device_handler(device_request)
                if fs:
                    device_actions_futures.extend(fs)
            if self.display and resp.screen_out.data:
                system_browser = browser_helpers.system_browser
                system_browser.display(resp.screen_out.data)

        if len(device_actions_futures):
            logging.info('Waiting for device executions to complete.')
            concurrent.futures.wait(device_actions_futures)

        logging.info('Finished playing assistant response.')
        self.conversation_stream.stop_playback()
        return continue_conversation

    def gen_assist_requests(self):
        """Yields: AssistRequest messages to send to the API."""

        config = embedded_assistant_pb2.AssistConfig(
            audio_in_config=embedded_assistant_pb2.AudioInConfig(
                encoding='LINEAR16',
                sample_rate_hertz=self.conversation_stream.sample_rate,
            ),
            audio_out_config=embedded_assistant_pb2.AudioOutConfig(
                encoding='LINEAR16',
                sample_rate_hertz=self.conversation_stream.sample_rate,
                volume_percentage=self.conversation_stream.volume_percentage,
            ),
            dialog_state_in=embedded_assistant_pb2.DialogStateIn(
                language_code=self.language_code,
                conversation_state=self.conversation_state,
                is_new_conversation=self.is_new_conversation,
            ),
            device_config=embedded_assistant_pb2.DeviceConfig(
                device_id=self.device_id,
                device_model_id=self.device_model_id,
            )
        )
        if self.display:
            config.screen_out_config.screen_mode = PLAYING
        # Continue current conversation with later requests.
        self.is_new_conversation = False
        # The first AssistRequest must contain the AssistConfig
        # and no audio data.
        yield embedded_assistant_pb2.AssistRequest(config=config)
        for data in self.conversation_stream:
            # Subsequent requests need audio data, but not config.
            yield embedded_assistant_pb2.AssistRequest(audio_in=data)


@click.command()
@click.option('--api-endpoint', default=ASSISTANT_API_ENDPOINT,
              metavar='<api endpoint>', show_default=True,
              help='Address of Google Assistant API service.')
@click.option('--credentials',
              metavar='<credentials>', show_default=True,
              default=os.path.join(click.get_app_dir('google-oauthlib-tool'),
                                   'credentials.json'),
              help='Path to read OAuth2 credentials.')
@click.option('--project-id',
              metavar='<project id>',
              help=('Google Developer Project ID used for registration '
                    'if --device-id is not specified'))
@click.option('--device-model-id',
              metavar='<device model id>',
              help=(('Unique device model identifier, '
                     'if not specifed, it is read from --device-config')))
@click.option('--device-id',
              metavar='<device id>',
              help=(('Unique registered device instance identifier, '
                     'if not specified, it is read from --device-config, '
                     'if no device_config found: a new device is registered '
                     'using a unique id and a new device config is saved')))
@click.option('--device-config', show_default=True,
              metavar='<device config>',
              default=os.path.join(
                  click.get_app_dir('googlesamples-assistant'),
                  'device_config.json'),
              help='Path to save and restore the device configuration')
@click.option('--lang', show_default=True,
              metavar='<language code>',
              default='en-US',
              help='Language code of the Assistant')
@click.option('--display', is_flag=True, default=False,
              help='Enable visual display of Assistant responses in HTML.')
@click.option('--verbose', '-v', is_flag=True, default=False,
              help='Verbose logging.')
@click.option('--input-audio-file', '-i',
              metavar='<input file>',
              help='Path to input audio file. '
              'If missing, uses audio capture')
@click.option('--output-audio-file', '-o',
              metavar='<output file>',
              help='Path to output audio file. '
              'If missing, uses audio playback')
@click.option('--audio-sample-rate',
              default=audio_helpers.DEFAULT_AUDIO_SAMPLE_RATE,
              metavar='<audio sample rate>', show_default=True,
              help='Audio sample rate in hertz.')
@click.option('--audio-sample-width',
              default=audio_helpers.DEFAULT_AUDIO_SAMPLE_WIDTH,
              metavar='<audio sample width>', show_default=True,
              help='Audio sample width in bytes.')
@click.option('--audio-iter-size',
              default=audio_helpers.DEFAULT_AUDIO_ITER_SIZE,
              metavar='<audio iter size>', show_default=True,
              help='Size of each read during audio stream iteration in bytes.')
@click.option('--audio-block-size',
              default=audio_helpers.DEFAULT_AUDIO_DEVICE_BLOCK_SIZE,
              metavar='<audio block size>', show_default=True,
              help=('Block size in bytes for each audio device '
                    'read and write operation.'))
@click.option('--audio-flush-size',
              default=audio_helpers.DEFAULT_AUDIO_DEVICE_FLUSH_SIZE,
              metavar='<audio flush size>', show_default=True,
              help=('Size of silence data in bytes written '
                    'during flush operation'))
@click.option('--grpc-deadline', default=DEFAULT_GRPC_DEADLINE,
              metavar='<grpc deadline>', show_default=True,
              help='gRPC deadline in seconds')
@click.option('--once', default=False, is_flag=True,
              help='Force termination after a single conversation.')
def main(api_endpoint, credentials, project_id,
         device_model_id, device_id, device_config,
         lang, display, verbose,
         input_audio_file, output_audio_file,
         audio_sample_rate, audio_sample_width,
         audio_iter_size, audio_block_size, audio_flush_size,
         grpc_deadline, once, *args, **kwargs):
    """Samples for the Google Assistant API.

    Examples:
      Run the sample with microphone input and speaker output:

        $ python -m googlesamples.assistant

      Run the sample with file input and speaker output:

        $ python -m googlesamples.assistant -i <input file>

      Run the sample with file input and output:

        $ python -m googlesamples.assistant -i <input file> -o <output file>
    """
    # Setup logging.
    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)

    # Load OAuth 2.0 credentials.
    try:
        with open(credentials, 'r') as f:
            credentials = google.oauth2.credentials.Credentials(token=None,
                                                                **json.load(f))
            http_request = google.auth.transport.requests.Request()
            credentials.refresh(http_request)
    except Exception as e:
        logging.error('Error loading credentials: %s', e)
        logging.error('Run google-oauthlib-tool to initialize '
                      'new OAuth 2.0 credentials.')
        sys.exit(-1)

    # Create an authorized gRPC channel.
    grpc_channel = google.auth.transport.grpc.secure_authorized_channel(
        credentials, http_request, api_endpoint)
    logging.info('Connecting to %s', api_endpoint)

    # Configure audio source and sink.
    audio_device = None
    if input_audio_file:
        audio_source = audio_helpers.WaveSource(
            open(input_audio_file, 'rb'),
            sample_rate=audio_sample_rate,
            sample_width=audio_sample_width
        )
    else:
        audio_source = audio_device = (
            audio_device or audio_helpers.SoundDeviceStream(
                sample_rate=audio_sample_rate,
                sample_width=audio_sample_width,
                block_size=audio_block_size,
                flush_size=audio_flush_size
            )
        )
    if output_audio_file:
        audio_sink = audio_helpers.WaveSink(
            open(output_audio_file, 'wb'),
            sample_rate=audio_sample_rate,
            sample_width=audio_sample_width
        )
    else:
        audio_sink = audio_device = (
            audio_device or audio_helpers.SoundDeviceStream(
                sample_rate=audio_sample_rate,
                sample_width=audio_sample_width,
                block_size=audio_block_size,
                flush_size=audio_flush_size
            )
        )
    # Create conversation stream with the given audio source and sink.
    conversation_stream = audio_helpers.ConversationStream(
        source=audio_source,
        sink=audio_sink,
        iter_size=audio_iter_size,
        sample_width=audio_sample_width,
    )

    if not device_id or not device_model_id:
        try:
            with open(device_config) as f:
                device = json.load(f)
                device_id = device['id']
                device_model_id = device['model_id']
                logging.info("Using device model %s and device id %s",
                             device_model_id,
                             device_id)
        except Exception as e:
            logging.warning('Device config not found: %s' % e)
            logging.info('Registering device')
            if not device_model_id:
                logging.error('Option --device-model-id required '
                              'when registering a device instance.')
                sys.exit(-1)
            if not project_id:
                logging.error('Option --project-id required '
                              'when registering a device instance.')
                sys.exit(-1)
            device_base_url = (
                'https://%s/v1alpha2/projects/%s/devices' % (api_endpoint,
                                                             project_id)
            )
            device_id = str(uuid.uuid1())
            payload = {
                'id': device_id,
                'model_id': device_model_id,
                'client_type': 'SDK_SERVICE'
            }
            session = google.auth.transport.requests.AuthorizedSession(
                credentials
            )
            r = session.post(device_base_url, data=json.dumps(payload))
            if r.status_code != 200:
                logging.error('Failed to register device: %s', r.text)
                sys.exit(-1)
            logging.info('Device registered: %s', device_id)
            pathlib.Path(os.path.dirname(device_config)).mkdir(exist_ok=True)
            with open(device_config, 'w') as f:
                json.dump(payload, f)

    device_handler = device_helpers.DeviceRequestHandler(device_id)

    @device_handler.command('action.devices.commands.OnOff')
    def onoff(on):
        if on:
            logging.info('Turning device on')
        else:
            logging.info('Turning device off')

    @device_handler.command('com.example.commands.BlinkLight')
    def blink(speed, number):
        logging.info('Blinking device %s times.' % number)
        delay = 1
        if speed == "SLOWLY":
            delay = 2
        elif speed == "QUICKLY":
            delay = 0.5
        for i in range(int(number)):
            logging.info('Device is blinking.')
            time.sleep(delay)

    with SampleAssistant(lang, device_model_id, device_id,
                         conversation_stream, display,
                         grpc_channel, grpc_deadline,
                         device_handler) as assistant:
        # If file arguments are supplied:
        # exit after the first turn of the conversation.
        if input_audio_file or output_audio_file:
            assistant.assist()
            return

        # If no file arguments supplied:
        # keep recording voice requests using the microphone
        # and playing back assistant response using the speaker.
        # When the once flag is set, don't wait for a trigger. Otherwise, wait.
        wait_for_user_trigger = not once
        while True:
            if wait_for_user_trigger:
                click.pause(info='Press Enter to send a new request...')
            continue_conversation = assistant.assist()
            # wait for user trigger if there is no follow-up turn in
            # the conversation.
            wait_for_user_trigger = not continue_conversation

            # If we only want one conversation, break.
            if once and (not continue_conversation):
                break


if __name__ == '__main__':
    main()



import os, sys
try:
    from setproctitle import setproctitle
except ImportError:
    setproctitle = None

from . import daemon
import time
from .. import pathtool
from ..termcolor import tc
if os.name == "nt":
    import win32serviceutil

class Service:
    def __init__ (self, name, working_dir, lockpath = None, win32service = None):
        self.name = name
        self.working_dir = os.path.normpath (working_dir)
        self.lockpath = lockpath or working_dir
        self.win32service = win32service
        setproctitle and setproctitle (name)
        pathtool.mkdir (working_dir)
        if lockpath:
            pathtool.mkdir (lockpath)
        os.chdir (working_dir)

    def start (self):
        if os.name == "nt":
            set_service_config (['start'])
        else:
            from . import Daemonizer
            if not Daemonizer (self.working_dir, self.name, lockpath = self.lockpath).runAsDaemon ():
                pid = daemon.status (self.lockpath, self.name)
                print ("{} {}".format (tc.debug (self.name), tc.error ("[already running:{}]".format (pid))))
                sys.exit ()

    def stop (self):
        if os.name == "nt":
            set_service_config (['stop'])
        else:
            daemon.kill (self.lockpath, self.name, True)

    def status (self, verbose = True):
        pid = daemon.status (self.lockpath, self.name)
        if verbose:
            if pid:
                print ("{} {}".format (tc.debug (self.name), tc.warn ("[running:{}]".format (pid))))
            else:
                print ("{} {}".format (tc.debug (self.name), tc.secondary ("[stopped]")))
        return pid

    def execute (self, cmd, fallback_user = None):
        if cmd == "stop":
            self.stop ()
            return False
        elif cmd == "status":
            self.status ()
            return False
        elif cmd == "start":
            self.start ()
        elif cmd == "restart":
            self.stop ()
            time.sleep (2)
            self.start ()
        elif cmd == "install":
            self.install (fallback_user)
            return False
        elif cmd == "update":
            self.update (fallback_user)
            return False
        elif cmd == "remove":
            self.remove ()
            return False
        else:
            raise AssertionError ('unknown command: %s' % cmd)

        if os.name == "nt":
            return False
        return True

    if os.name == "nt":
        def set_service_config (self, argv = []):
            argv.insert (0, "")
            script = os.path.join (os.getcwd (), sys.argv [0])
            win32serviceutil.HandleCommandLine(self.win32service, "%s.%s" % (script [:-3], self.win32service.__name__), argv)

        def install (self):
            self.set_service_config (['--startup', 'auto', 'install'])

        def remove (self):
            self.set_service_config (['remove'])

        def update (self):
            self.set_service_config (['update'])

    else:
        def install (self, fallback_user, update = False):
            name = self.name.split ('/')[-1]
            service_script = '/etc/systemd/system/{}.service'.format (name)
            if not update and os.path.exists (service_script):
                raise SystemError ('service {} is already exists. use update command'.format (name))
            if fallback_user:
                user = 'root'
            else:
                user = os.getenv("SUDO_USER") or os.getenv ("LOGNAME")
            args = sys.argv [1:]
            if args and hasattr (self, args [-1]):
                args = args [:-1]

            if self.name == 'skitai/smtpda':
                import skitai
                serve = os.path.join (os.path.dirname (skitai.__file__), 'scripts', 'commands', 'smtpda.py')
            else:
                serve = os.path.abspath (sys.argv [0])

            script = SERVICE_TEMPLATE.format (
                name = name.title (),
                user = user,
                executable = sys.executable,
                serve = serve,
                working_dir = self.working_dir,
                pidfile = name,
                args = " ".join (['"{}"'.format (each) for each in args])
            )

            print ('# {}.service =========================================\n'.format (name))
            print (script)
            print ('# end of {}.service ==================================\n'.format (name))

            try:
                with open (service_script, 'w') as f:
                    f.write (script)

            except PermissionError:
                service_script = os.path.join (os.getcwd (), '{}.service'.format (name))
                with open (service_script, 'w') as f:
                    f.write (script)
                os.system ('sudo mv {} /etc/systemd/system'.format (service_script))

            sudo = ''
            if os.getuid() != 0:
                sudo = 'sudo '
            os.system ("{}systemctl daemon-reload".format (sudo))
            os.system ("{}systemctl enable {}".format (sudo, name))

            print ("service {} installed, to start service,".format (name))
            print ("  sudo systemctl start {}".format (name))

        def update (self, fallback_user):
            self.install (fallback_user, True)

        def remove (self):
            name = self.name.split ('/')[-1]
            service_script = '/etc/systemd/system/{}.service'.format (name)
            if not os.path.exists (service_script):
                return
            os.remove (service_script)

    uninstall = remove

SERVICE_TEMPLATE = '''[Unit]
Description=Skitai {name} Service
After=syslog.target network.target

[Service]
Type=simple
WorkingDirectory={working_dir}
User={user}
Group={user}
StandardOutput=syslog
StandardError=syslog
Restart=on-failure
ExecStart={executable} {serve} {args}
ExecReload=/bin/kill -s HUP $MAINPID
RestartSec=2s
TimeoutStopSec=5
KillMode=mixed

[Install]
WantedBy=multi-user.target
'''




# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _snowboydetect
else:
    import _snowboydetect

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


class SnowboyDetect(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, resource_filename, model_str):
        _snowboydetect.SnowboyDetect_swiginit(self, _snowboydetect.new_SnowboyDetect(resource_filename, model_str))

    def Reset(self):
        return _snowboydetect.SnowboyDetect_Reset(self)

    def RunDetection(self, *args):
        return _snowboydetect.SnowboyDetect_RunDetection(self, *args)

    def SetSensitivity(self, sensitivity_str):
        return _snowboydetect.SnowboyDetect_SetSensitivity(self, sensitivity_str)

    def SetHighSensitivity(self, high_sensitivity_str):
        return _snowboydetect.SnowboyDetect_SetHighSensitivity(self, high_sensitivity_str)

    def GetSensitivity(self):
        return _snowboydetect.SnowboyDetect_GetSensitivity(self)

    def SetAudioGain(self, audio_gain):
        return _snowboydetect.SnowboyDetect_SetAudioGain(self, audio_gain)

    def UpdateModel(self):
        return _snowboydetect.SnowboyDetect_UpdateModel(self)

    def NumHotwords(self):
        return _snowboydetect.SnowboyDetect_NumHotwords(self)

    def ApplyFrontend(self, apply_frontend):
        return _snowboydetect.SnowboyDetect_ApplyFrontend(self, apply_frontend)

    def SampleRate(self):
        return _snowboydetect.SnowboyDetect_SampleRate(self)

    def NumChannels(self):
        return _snowboydetect.SnowboyDetect_NumChannels(self)

    def BitsPerSample(self):
        return _snowboydetect.SnowboyDetect_BitsPerSample(self)
    __swig_destroy__ = _snowboydetect.delete_SnowboyDetect

# Register SnowboyDetect in _snowboydetect:
_snowboydetect.SnowboyDetect_swigregister(SnowboyDetect)

class SnowboyVad(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, resource_filename):
        _snowboydetect.SnowboyVad_swiginit(self, _snowboydetect.new_SnowboyVad(resource_filename))

    def Reset(self):
        return _snowboydetect.SnowboyVad_Reset(self)

    def RunVad(self, *args):
        return _snowboydetect.SnowboyVad_RunVad(self, *args)

    def SetAudioGain(self, audio_gain):
        return _snowboydetect.SnowboyVad_SetAudioGain(self, audio_gain)

    def ApplyFrontend(self, apply_frontend):
        return _snowboydetect.SnowboyVad_ApplyFrontend(self, apply_frontend)

    def SampleRate(self):
        return _snowboydetect.SnowboyVad_SampleRate(self)

    def NumChannels(self):
        return _snowboydetect.SnowboyVad_NumChannels(self)

    def BitsPerSample(self):
        return _snowboydetect.SnowboyVad_BitsPerSample(self)
    __swig_destroy__ = _snowboydetect.delete_SnowboyVad

# Register SnowboyVad in _snowboydetect:
_snowboydetect.SnowboyVad_swigregister(SnowboyVad)






from ... import siesta
import json
import base64
import librosa
from io import BytesIO
import soundfile
import os
from ... import pathtool

def resampling (audioFilePath):
    y, sr = librosa.load (audioFilePath, sr = 22050, mono = True)
    with BytesIO () as f:
       soundfile.write (f, y, 16000, format = "WAV")       
       return f.getvalue ()
   
openApiURL = "http://aiopen.etri.re.kr:8000"
API = siesta.API (openApiURL)

def get_speech (audioFilePath, lang = "korean"):
    from . import ACCESS_KEY
    
    audioContents = base64.b64encode (resampling (audioFilePath)).decode("utf8")
    requestJson = {
        "access_key": ACCESS_KEY,
        "argument": {"audio": audioContents, "language_code": lang},
        "request_id": "sns",
    }
    resp = API.WiseASR.Recognition.post (
        requestJson,
        {"Content-Type": "application/json; charset=UTF-8"}     
    )
    try:
        assert resp.status_code == 200
        text = resp.data ['return_object']['recognized']
    except (AssertionError, KeyError):
        return ''
    return text



# pip3 install awscli --upgrade --user
# aws configure

import boto3

ec2 = boto3.resource ('ec2')
ec2cli = boto3.client ('ec2')
elb = boto3.client('elb')
cw = boto3.client('cloudwatch')
elb2 = boto3.client('elbv2')
autoscaling = boto3.client('autoscaling')
rt53 = boto3.client('route53')


# Hans Roh (Apr 5, 2020)

import time
import sys, os
import fabric
from rs4.apis.aws import ec2, elb2, ec2cli, rt53, autoscaling
import threading
import requests
from pprint import pprint
import botocore
from urllib.parse import urlparse

def extract_domain(url):
    parsed_domain = urlparse(url)
    domain = parsed_domain.netloc or parsed_domain.path # Just in case, for urls without scheme
    domain_parts = domain.split('.')
    if len(domain_parts) > 2:
        return '.'.join(domain_parts[-(2 if domain_parts[-1] in {
            'com', 'net', 'org', 'io', 'ly', 'me', 'sh', 'fm', 'us'} else 3):])
    return domain


class Resources:
    def __init__ (self, system, initial_image, key_name, subnets, security_groups):
        self.system = system

        self.KEY_NAME = key_name
        self.INITIAL_AMI_ID = initial_image
        self.AMI_ID = initial_image
        self.SUBNETS = subnets
        self.SECURITY_GROUPS = security_groups

        requisits = self.get_images ()
        if requisits:
            self.AMI_ID = requisits [0]['ImageId']
        self.LAUNCH_TEMPLATE = '{}-launch-template'.format (self.system)
        self.AUTO_SCALING_GROUP = '{}-as-group'.format (self.system)
        self.EC2_NAME = '{}-elb-member'.format (self.system)
        self.ELB = '{}-elb'.format (self.system)
        self.ELB_TARGET = '{}-elb-target'.format (self.system)

        self.ELB_GROUP_ARN = None
        self.ELB_ARN = None
        self.NEW_ELB = False
        self.VPC_ID = None
        self.LAUNCH_TEMPLATE_VERSION = "1"

    # getting resource ------------------------------------------
    def get_images (self):
        images = ec2cli.describe_images (Owners=['self'])['Images']
        images = [img for img in images if img ['Name'].startswith ('{}-'.format (self.system))]
        if not images:
            return []
        images.sort (key = lambda x: x ['Name'], reverse = True)
        return images

    def get_latest_template_version (self):
        templates = ec2cli.describe_launch_template_versions (LaunchTemplateName = self.LAUNCH_TEMPLATE, MaxResults = 1)
        return templates ['LaunchTemplateVersions'][0]

    # resources ---------------------------------------------------
    def create_load_balancer_if_not_exists (self):
        print ('checking elastic load balancer...')
        elbs = [e for e in elb2.describe_load_balancers () ["LoadBalancers"] if e ['LoadBalancerName'] == self.ELB]
        if not elbs:
            self.NEW_ELB = True
            elbs = elb2.create_load_balancer (
                Name = self.ELB,
                Subnets = ['subnet-94c348d8', 'subnet-b76d8ecc'],
                SecurityGroups = self.SECURITY_GROUPS,
                Tags=[{'Key': 'system', 'Value': self.system }],
                Type='application',
                Scheme='internet-facing',
                IpAddressType='ipv4'
            ) ['LoadBalancers']
        self.ELB_ARN = elbs [0]['LoadBalancerArn']
        self.ELB_DNS_NAME = elbs [0]['DNSName']
        self.VPC_ID = elbs [0] ['VpcId']

    def add_listener_if_not_exists (self, cert, target_group_arns = None):
        listensers = elb2.describe_listeners (LoadBalancerArn = self.ELB_ARN) ['Listeners']
        protocols = [ lsn ['Protocol'] for lsn in listensers ]
        target_group_arns = target_group_arns or [(self.ELB_GROUP_ARN, 1)]
        target_groups = [{'TargetGroupArn': target_arn, 'Weight': weight} for target_arn, weight in target_group_arns]
        if 'HTTPS' not in protocols:
            elb2.create_listener (
                LoadBalancerArn = self.ELB_ARN, Protocol = 'HTTPS', Port = 443,
                DefaultActions = [{'ForwardConfig': {'TargetGroups': target_groups}, 'Type': 'forward'}],
                Certificates = [ {'CertificateArn': cert} ],
                SslPolicy='ELBSecurityPolicy-2016-08'
            )

        if 'HTTP' not in protocols:
            elb2.create_listener (
                LoadBalancerArn = self.ELB_ARN, Protocol = 'HTTP', Port = 80,
                DefaultActions = [{
                    'Type': 'redirect',
                    "RedirectConfig": {
                        "Protocol": "HTTPS",
                        "Port": "443",
                        "StatusCode": "HTTP_301",
                    }
                }]
            )

    def create_launch_template_if_not_exists (self, instance_type, instance_role_arn):
        print ('checking launch template...')
        tps = [e for e in ec2cli.describe_launch_templates () ["LaunchTemplates"] if e ['LaunchTemplateName'] == self.LAUNCH_TEMPLATE]
        if not tps:
            ec2cli.create_launch_template (
                LaunchTemplateName = self.LAUNCH_TEMPLATE,
                LaunchTemplateData = {
                    'CreditSpecification': {'CpuCredits': 'standard'},
                    'DisableApiTermination': False,
                    'IamInstanceProfile': {'Arn': instance_role_arn},
                    'ImageId': self.AMI_ID,
                    'InstanceType': instance_type,
                    'KeyName': self.KEY_NAME,
                    'Monitoring': {'Enabled': False},
                    'SecurityGroupIds': self.SECURITY_GROUPS,
                    'TagSpecifications': [{'ResourceType': 'instance',
                                            'Tags': [
                                                {'Key': 'system', 'Value': self.system},
                                                {'Key': 'Name', 'Value': self.EC2_NAME}
                                            ]
                                        }]
                    }
            )

    def create_target_group_if_not_exists (self, target_group_name = None):
        print ('checking target group...')
        is_default_target_group = target_group_name is None
        target_group_name = target_group_name or self.ELB_TARGET
        tgs = [e for e in elb2.describe_target_groups () ["TargetGroups"] if e ['TargetGroupName'] == target_group_name]
        if not tgs:
            tgs = elb2.create_target_group (
                Name = target_group_name,
                VpcId = self.VPC_ID,
                Protocol = 'HTTP',
                Port = 80
            ) ['TargetGroups']

        if is_default_target_group:
            # default
            self.ELB_GROUP_ARN = tgs [0]['TargetGroupArn']
        return tgs [0]['TargetGroupArn']

    def create_auto_scaling_group_if_not_exist (self, min_, max_, auto_scaling_group_name = None, target_group_arn = None):
        print ('checking auto scaling group...')
        auto_scaling_group_name = auto_scaling_group_name or self.AUTO_SCALING_GROUP
        asg = [e for e in autoscaling.describe_auto_scaling_groups () ["AutoScalingGroups"] if e ['AutoScalingGroupName'] == auto_scaling_group_name]
        if not asg:
            autoscaling.create_auto_scaling_group (
                AutoScalingGroupName = auto_scaling_group_name,
                MinSize = min_,
                MaxSize = max_,
                VPCZoneIdentifier = ','.join (self.SUBNETS),
                LaunchTemplate = {'LaunchTemplateName': self.LAUNCH_TEMPLATE, 'Version': '$Latest'},
                TargetGroupARNs = [target_group_arn or self.ELB_GROUP_ARN],
                TerminationPolicies = ['OldestLaunchTemplate', 'OldestInstance'],
                Tags = [{'Key': 'system',
                    'PropagateAtLaunch': True,
                    'ResourceId': auto_scaling_group_name,
                    'ResourceType': 'auto-scaling-group',
                    'Value': self.system
                }]
            )

    def add_scaling_policy (self, name, metric_type = 'ASGAverageCPUUtilization', value = 75.0, auto_scaling_group_name = None):
        auto_scaling_group_name = auto_scaling_group_name or self.AUTO_SCALING_GROUP
        autoscaling.put_scaling_policy (
            AutoScalingGroupName = auto_scaling_group_name,
            PolicyName = name,
            PolicyType = 'TargetTrackingScaling',
            TargetTrackingConfiguration = {'DisableScaleIn': False,
                                            'PredefinedMetricSpecification': {'PredefinedMetricType': metric_type},
                                            'TargetValue': value}
        )

    def launch_instance (self, instance_type, user_data = None, dryrun = False):
        instances = ec2.create_instances (
            ImageId = self.INITIAL_AMI_ID,
            MinCount = 1,
            MaxCount = 1,
            InstanceType = instance_type,
            Monitoring = {'Enabled': False},
            SecurityGroupIds = self.SECURITY_GROUPS,
            DryRun = dryrun,
            KeyName = self.KEY_NAME,
            UserData = isinstance (user_data, str) and user_data or '\n'.join (user_data),
            InstanceInitiatedShutdownBehavior = 'terminate',
            TagSpecifications = [{
                    'ResourceType': 'instance',
                    'Tags': [
                            {'Key': 'Name', 'Value': '{}-elb-member'.format (self.system)},
                            {'Key': 'system', 'Value': self.system},
                    ]
                },
            ])

        inst = instances [0]
        print ('creating instance...: {} {}'.format (inst.id, inst.public_dns_name))
        for i in range (20):
            inst.reload()
            state = inst.state ['Name']
            if state == 'running':
                print ('instance created')
                return inst
            print ('- instance state: {}'.format (state))
            time.sleep (6)
        inst.terminate ()
        raise SystemError ('instance creation timeout')

    def create_image (self, inst, testfunc):
        image = None
        print ('waiting for app...')
        time.sleep (200)
        for i in range (60):
            if testfunc (self, inst):
                image = inst.create_image (Name = '{}-{}'.format (self.system, time.strftime ('%y%m%d%H%M', time.localtime (time.time ()))))
                break
            print ('- waiting...')
            time.sleep (10)

        if not image:
            raise SystemError ('instance activation failed')

        print ('creating image...')
        time.sleep (60)
        for i in range (30):
            image.reload ()
            if image.state == 'available':
                break
            print ('- image state: {}'.format (image.state))
            time.sleep (30)
        print ('image created: {}'.format (image.id))
        self.remove_old_images (3)
        assert testfunc (self, inst)
        return image

    def update_template (self, image):
        print ('applying {} to template...'.format (image.id))
        latest = self.get_latest_template_version ()
        ec2cli.create_launch_template_version (
            LaunchTemplateName = self.LAUNCH_TEMPLATE,
            SourceVersion = str (latest ['VersionNumber']),
            LaunchTemplateData = {'ImageId': image.id}
        )
        latest = self.get_latest_template_version ()
        print ('new launch template version created: {}'.format (latest ['VersionNumber']))

    def cleanup (self):
        print ('cleanup load balancer...')
        elbs = [e for e in elb2.describe_load_balancers () ["LoadBalancers"] if e ['LoadBalancerName'] == self.ELB]
        if elbs:
            elb2.delete_load_balancer (LoadBalancerArn=elbs [0]['LoadBalancerArn'])

        print ('cleanup auto scaler...')
        if autoscaling.describe_auto_scaling_groups () ["AutoScalingGroups"]:
            autoscaling.delete_auto_scaling_group (AutoScalingGroupName = self.AUTO_SCALING_GROUP, ForceDelete = True)
            time.sleep (10.)

        print ('cleanup target group...')
        tgs = [e for e in elb2.describe_target_groups () ["TargetGroups"] if e ['TargetGroupName'] == self.ELB_TARGET]
        if tgs:
            elb2.delete_target_group (TargetGroupArn = tgs [0]['TargetGroupArn'])

        print ('cleanup instances...')
        for inst in ec2.instances.filter(Filters=[{'Name': 'tag:Name', 'Values': [self.EC2_NAME]}]):
            if inst.state ['Name'] == 'running':
                print ('- terminating instance: {}'.format (inst.id))
                inst.terminate ()


class Deploy (Resources):
    # utils --------------------------------------------------
    def ssh (self, inst, key_file, user = 'ubuntu'):
        if isinstance (inst, str):
            inst = ec2.Instance (inst)
        if inst.state ['Name'] != 'running':
            raise SystemError ('instance not running')
        return fabric.Connection (inst.public_dns_name, user, connect_kwargs = dict (key_filename = key_file))

    def ssh_with_password (self, host, user, password, port = 22):
        return fabric.Connection (host, user, port = port, connect_kwargs = dict (password = password))

    def set_auto_scaling_group_range (self, min_ = None, max_ = None):
        return autoscaling.update_auto_scaling_group (
            AutoScalingGroupName = self.AUTO_SCALING_GROUP,
            MinSize = min_,
            MaxSize = max_
        )

    def get_hosted_zone_id (self, domain):
        root_domain = extract_domain (domain)
        zone = [ each for each in rt53.list_hosted_zones () ['HostedZones'] if each ['Name'] == root_domain + '.' ][0]
        return zone ['Id'].split ('/') [-1]

    def set_dns (self, type, domain, value):
        hosted_zone = self.get_hosted_zone_id (domain)
        rds = rt53.list_resource_record_sets (HostedZoneId = hosted_zone) ['ResourceRecordSets']
        record = None
        for rd in rds:
            # alias_zone_id
            if rd ['Name'] == domain + '.' and rd ['Type'] == type:
                record = rd

        alias_zone_id = None
        if record and 'AliasTarget' in record:
            alias_zone_id = record ['AliasTarget']['HostedZoneId']

        change_batch = {
            'Comment': 'Create/Update ELB dns entry',
            'Changes': [{
                'Action': 'UPSERT',
                'ResourceRecordSet': {'Name': '{}.'.format (domain), 'Type': type}
            }]
        }
        if alias_zone_id:
            change_batch ['Changes'][0]['ResourceRecordSet']['AliasTarget'] = { 'HostedZoneId': alias_zone_id, 'DNSName': value, 'EvaluateTargetHealth': False}
        else:
            change_batch ['Changes'][0]['ResourceRecordSet']['ResourceRecords'] = [{'Value': value}]
            change_batch ['Changes'][0]['ResourceRecordSet']['TTL'] = 300
        rt53.change_resource_record_sets(
            HostedZoneId = hosted_zone,
            ChangeBatch = change_batch
        )

    def link_elb_to_domain (self, domain):
        if not self.NEW_ELB:
            return
        self.set_dns ('A', domain, self.ELB_DNS_NAME)

    def reassociate_eip (self, name, inst = None):
        eip = ec2cli.describe_addresses(Filters=[ {'Name': 'domain', 'Values': ['vpc']}, {'Name': 'tag:Name', 'Values': [name]} ])['Addresses'] [0]
        alloc_id = eip ['AllocationId']
        if 'AssociationId' in eip:
            assoc_id = eip ['AssociationId']
            ec2cli.disassociate_address (AssociationId = assoc_id)
        if inst:
            ec2cli.associate_address (AllocationId = alloc_id, InstanceId=inst.id)

    def rescale (self, image, current_count, desired_count):
        def runnings ():
            return [ deployed.id for deployed in self.get_instances_by_image (image, 'running') ]

        new_instances = runnings ()
        valids = [ t ['Target']['Id'] for t in self.get_elb_targets () if t ['Target']['Id'] in new_instances and t ['TargetHealth']['State'] != 'draining' ]
        exists_new_instances = len (valids)

        self.set_auto_scaling_group_range (current_count + desired_count, current_count + desired_count)
        print ('desired new isntance count: {}'.format (desired_count))
        for i in range (20):
            new_instances = runnings ()
            print ('- found {} new versioned instances, target is {}'.format (len (new_instances), exists_new_instances + desired_count))
            if len (new_instances) >= (exists_new_instances + desired_count) * 0.75:
                print ('{} new running instances fulfill minimum requirements'.format (len (new_instances)))
                break
            time.sleep (30)

        for i in range (20):
            new_instances = runnings ()
            valids = [ t ['Target']['Id'] for t in self.get_elb_targets () if t ['Target']['Id'] in new_instances and t ['TargetHealth']['State'] in ('healthy', 'draining') ]
            print ('- healthy / new instances: {} / {}'.format (len (valids), len (new_instances)))
            if len (valids) > len (new_instances) * 0.75:
                print ('{} healthy instances fulfill minimum requirements'.format (len (valids)))
                break
            time.sleep (30)

        self.set_auto_scaling_group_range (current_count, current_count)
        return new_instances

    def register_target (self, inst, target_group_arn = None):
        print ('registering new elb2 member: {}'.format (inst.id))
        elb2.register_targets (
            TargetGroupArn = target_group_arn or self.ELB_GROUP_ARN,
            Targets = [{'Id': inst.id, 'Port': port}]
        )

    def degister_target (self, inst, target_group_arn = None):
        print ('deregistering new elb2 member: {}'.format (inst.id))
        elb2.deregister_targets (
            TargetGroupArn = target_group_arn or self.ELB_GROUP_ARN,
            Targets = [{'Id': inst.id, 'Port': port}]
        )

    def remove_image (self, image_id):
        ec2cli.deregister_image(ImageId = image_id)
        owner_id = self.get_images () [0]['OwnerId']
        for s in ec2cli.describe_snapshots (OwnerIds = [owner_id]) ['Snapshots']:
            if s ['Description'].find (image_id) != -1:
                ec2cli.delete_snapshot (SnapshotId=s['SnapshotId'])
                break

    def remove_old_images (self, keep = 3):
        images = self.get_images ()
        print ('unregistering images...')
        for img in images [keep:]:
            self.remove_image (img ['ImageId'])
            print ('- image unregistered: {}'.format (id))

    def get_elb_targets (self, target_group_arn = None):
        return elb2.describe_target_health (TargetGroupArn = target_group_arn or self.ELB_GROUP_ARN) ['TargetHealthDescriptions']

    def filter_state (self, lst, state):
        if not state:
            return lst
        if isinstance (state, str):
            state = [state]
        return [inst for inst in lst if inst.state ['Name'] in state]

    def get_instances_by_name (self, name, state = None):
        return self.filter_state (list (ec2.instances.filter(Filters=[{'Name': 'tag:Name', 'Values': [name]}])), state)

    def get_instances_by_image (self, image, state = None):
        return self.filter_state (list (ec2.instances.filter(Filters=[{'Name': 'image-id', 'Values': [image.id]}])), state)

    def get_old_instances_by_image (self, new_image):
        news = [inst.id for inst in self.get_instances_by_image (new_image, 'running')]
        valids = [ t ['Target']['Id'] for t in self.get_elb_targets () if t ['TargetHealth']['State'] != 'draining' ]
        return [ inst for inst in ec2.instances.filter(Filters=[{'Name': 'tag:Name', 'Values': [self.EC2_NAME]}]) if inst.id not in news and inst.id in valids and inst.state ['Name'] == 'running']


if __name__ == '__main__':

    # This is just an example and will not work.

    import time
    import sys, os
    from rs4.apis.aws import ec2, autoscale_deploy
    import requests
    from pprint import pprint
    import math

    # resources ------------------------------------------
    INITIAL_AMI_ID = 'ami-0cd7b0de75f5a35d1'
    SUBNETS = ['subnet-94c', 'subnet-b76']
    SECURITY_GROUPS = ['sg-fa2bec90', 'sg-0ce0adcb']
    INSTANCE_ROLE_ARN = 'arn:aws:iam::'
    SSL_CERTIFICATION_ARN ='arn:aws:acm:'

    # config ------------------------------------------
    SYSTEM = 'sample'
    KEY_NAME = 'my-key'
    KEY_FILE = os.path.expanduser ('~/.aws/.aws-ap-northeast-2.pem')
    AUTO_SCALING_RANGE = (1, 3)
    USERDATA = [
        '#!/bin/bash',
        'cd /home/ubuntu',
        'apt update',
        'DEBIAN_FRONTEND=noninteractive apt upgrade -yq',
        'git clone https://gitlab+deploy-token-156019:asdf@gitlab.com',
        'git checkout -b master && git pull origin master',
        './resources/install.sh',
        'apt install nginx -y',
        'rm /etc/nginx/sites-enabled/default',
        'ln -s /home/ubuntu/gongbi/resources/etc/nginx.default.conf /etc/nginx/sites-enabled/default',
        'systemctl reload nginx && systemctl status nginx',
        'mkdir /var/log/skitai && chown -R ubuntu:ubuntu /var/log/skitai',
        'mkdir /var/tmp/skitai && chown -R ubuntu:ubuntu /var/tmp/skitai',
        'chown -R ubuntu:ubuntu /home/ubuntu',
        'su - ubuntu -c "/usr/local/bin/skitai smtpda install"',
        'su - ubuntu -c "/home/ubuntu/gongbi/serve.py install"',
        'systemctl start smtpda',
        'systemctl start gongbi && systemctl status gongbi'
    ]

    def testfunc (dep, inst):
        try:
            resp = requests.get ('http://{}'.format (inst.public_dns_name))
        except:
            return False
        if not (resp.status_code == 200 and 'mapVuexItems' in resp.text):
            return False
        with dep.ssh (inst, KEY_FILE) as ssh:
            if "Active: active (running)" not in ssh.run ('sudo systemctl status gongbi', hide = True).stdout:
                return False
            if "Active: active (running)" not in ssh.run ('sudo systemctl status smtpda', hide = True).stdout:
                return False
        return True


    if __name__ == '__main__':
        from rs4 import argopt
        argopt.add_option (None, '--id=INSTANCE_ID')
        argopt.add_option (None, '--clean')
        argopt.add_option ('-s=', '--strategy=STRATEGY', 'rescale|patch')
        argopt.add_option (None, '--system=SYSTEM_NAME')

        opts = argopt.collect ()
        if '--help' in opts:
            argopt.usage (True)

        dep = autoscale_deploy.AutoScaleDeploy (SYSTEM, INITIAL_AMI_ID, KEY_NAME, SUBNETS, SECURITY_GROUPS)
        ISNTANCE_ID = opts.any ('--id')
        STRATEGY = opts.any ('--strategy', '-s')
        if STRATEGY == 'rescale':
            if ISNTANCE_ID: # for simulating
                inst = ec2.Instance (ISNTANCE_ID)
                image = ec2.Image (dep.get_images () [0]['ImageId'])
            else:
                inst = dep.launch_instance ('t3.nano', USERDATA)
                image = dep.create_image (inst, testfunc)
                dep.update_template (image)
                inst.terminate ()

        if '--clean' in opts:
            dep.cleanup ()
            sys.exit ()

        if not STRATEGY:
            print ('\nERROR: --strategy required')
            argopt.usage (False)
            sys.exit (1)

        dep.create_load_balancer_if_not_exists ()
        dep.create_target_group_if_not_exists ()
        dep.add_listener_if_not_exists (SSL_CERTIFICATION_ARN)
        dep.create_launch_template_if_not_exists ('t3.nano', INSTANCE_ROLE_ARN)
        dep.create_auto_scaling_group_if_not_exist (3, 16)
        dep.add_scaling_policy ('cpu-75', 'ASGAverageCPUUtilization', 75.0)
        dep.link_elb_to_domain ('gongbi.co.kr')

        if STRATEGY == 'rescale':
            current_count = max (1, len ([ t ['Target']['Id'] for t in dep.get_elb_targets () if t ['TargetHealth']['State'] == 'healthy' ]))
            desired_count = max (1, math.ceil (current_count * 0.55))

            for i in range (3):
                print ('we want {} new versioning instances (session: {})'.format (desired_count, i + 1))
                new_instances = dep.rescale (image, current_count, desired_count)
                print ('created {} new version instances'.format (len (new_instances)))

                print ('waiting for starting instances draining...')
                time.sleep (10) #$ waiting for applying min, max policy
                for i in range (5):
                    drainings = [ t ['Target']['Id'] for t in dep.get_elb_targets () if t ['TargetHealth']['State'] == 'draining' ]
                    if drainings:
                        print ('{} old instances is on draining'.format (len (drainings)))
                        break
                    time.sleep (30)

                current_count = len ([ t ['Target']['Id'] for t in dep.get_elb_targets () if t ['TargetHealth']['State'] == 'healthy' ])
                desired_count = len (dep.get_old_instances_by_image (image))
                if not desired_count:
                    print ('old instances are completely removed')
                    break

            dep.set_auto_scaling_group_range (*AUTO_SCALING_RANGE)
            print ('autoscale range has been reinitialized.')

        elif STRATEGY == 'patch':
            targets = dep.get_elb_targets ()
            print ('pathcing {} instances...'.format (len (targets)))
            for target in dep.get_elb_targets ():
                instance_id = target ['Target']['Id']
                inst = ec2.Instance (instance_id)
                print ('- pathcing: {}'.format (inst.id))
                with dep.ssh (inst, KEY_FILE, 'ubuntu') as ssh:
                    ssh.run ('ls')





from cssselect import GenericTranslator
import lxml.etree
import lxml.html
import lxml.html.clean
import re
import traceback
from copy import deepcopy
from .. import strutil
from . import cssutil

TABSSPACE = re.compile(r'[\s\t]+')
SIMPLIFIED_CHILD_CSS = re.compile (r"\s*([<\[])(\-?)([0-9]+)[>\]]$")

def inner_trim(value):
	if strutil.is_str_like (value):
		# remove tab and white space
		value = re.sub(TABSSPACE, ' ', value)
		value = ''.join(value.splitlines())
		return value.strip()
	return ''

def remove_dquotes (s):
	s = s.strip ()
	if not s:
		return ""
	if s [0] in '\'"':		
		return s [1:-1].replace ('\%s' % s [0], s [0])
	return s		
	
def _filter_by_data (element, text):
	op, op2, text = text [0], text [1], remove_dquotes (text [2:])
	if op2 not in "=|~":
		text = op2 + text
		op2 = "="
		
	if op2 == "=":
		etext = element.text	
		if not etext:
			etext = ""
		else:	
			etext = etext.strip ()
	elif op2 == "|":
		etext = nops.get_text (element)
	else: # "~"		
		etext = element.tail and element.tail.strip () or ''	

	if op == "=":
		return etext == text
	elif op == "/":
		rx = re.compile (text, re.M|re.S)
		return rx.search (etext)
	elif op == "!":
		return etext != text
	elif op == "*":
		return etext.find (text) != -1
	elif op == "^":	
		return etext.startswith (text)
	elif op == "$":
		return etext.endswith (text)
	elif op == "~":	
		return text in etext.split ()
	else:
		raise AssertionError ("unknown operator %s" % op)	
	
def filter_by_data (elements, text):
	return [element for element in elements if _filter_by_data (element, text)]		
										
class nops:
	@classmethod
	def from_string (cls, html):
		if strutil.is_str_like (html) and html.startswith('<?'):
			html = re.sub(r'^\<\?.*?\?\>', '', html, flags=re.DOTALL)				
		try:
			return lxml.html.fromstring(html)
		except Exception:
			traceback.print_exc()
			return None
		
	@classmethod
	def to_string (cls, node, encoding = "utf8", method='html', doctype = None):
		return lxml.html.tostring (node, encoding = encoding, method = method, doctype = doctype).decode ("utf8")
	
	@classmethod
	def by_xpath (cls, node, expression):
		if expression.find ("re:test") != -1:
			return cls.by_xpath_re (node, expression)
		items = node.xpath(expression)
		return items
	
	@classmethod
	def by_xpath_re (cls, node, expression):
		regexp_namespace = "http://exslt.org/regular-expressions"
		items = node.xpath(expression, namespaces={'re': regexp_namespace})
		return items
	
	@classmethod
	def by_css (cls, node, selector):
		m = SIMPLIFIED_CHILD_CSS.search (selector)
		lev = ""
		index = 0
		if m:
			index = int (m.group (3))
			if m.group (2) == "-":
				index *= -1
			if m.group (1) == "<":
				lev = "child"				
			else:	
				lev = "queue"					
				
			selector = SIMPLIFIED_CHILD_CSS.sub ("", selector).strip ()

		if selector.strip ().startswith ("/"):
			founds = cls.by_xpath (node, selector)				
		else:
			founds = cls.by_xpath (node, GenericTranslator().css_to_xpath(selector))
			
		if not lev:	
			return founds
		
		if lev == "queue":		
			try:
				return [founds [index]]
			except IndexError:
				return []			
		
		r = []
		for each in founds:
			try:
				r.append (cls.get_children (each) [index])
			except IndexError:
				pass
		return r			
		
	@classmethod
	def by_id (cls, node, idd):
		selector = '//*[@id="%s"]' % idd
		elems = node.xpath(selector)
		if elems:
			return elems[0]
		return None

	@classmethod
	def by_tag_attr (cls, node, tag=None, attr=None, value=None, childs=False):
		NS = "http://exslt.org/regular-expressions"
		# selector = tag or '*'
		selector = 'descendant-or-self::%s' % (tag or '*')
		if attr and value:
			selector = '%s[re:test(@%s, "%s", "i")]' % (selector, attr, value)		
		elems = node.xpath(selector, namespaces={"re": NS})
		# remove the root node
		# if we have a selection tag
		if node in elems and (tag or childs):
			elems.remove(node)
		return elems
	
	@classmethod
	def by_tag (cls, node, tag):
		return cls.by_tags (node, [tag])
	
	@classmethod
	def by_tags (cls, node, tags):
		selector = ','.join(tags)
		elems = cls.by_css (node, selector)
		# remove the root node
		# if we have a selection tag
		if node in elems:
			elems.remove(node)
		return elems
	
	@classmethod
	def by_tags_nearest (cls, node, tags):
		if type (tags) is str:
			tags = tags.split (",")
		children = cls.get_children (node)
		while children:
			got = [child for child in children if child.tag in tags]
			if got: 
				return got
			children = cls.get_children (children [0])		
		return []
	
	@classmethod
	def by_tag_nearest (cls, node, tag):		
		return cls.by_tags_nearest (node, tag)
		
	@classmethod
	def by_lt (cls, node, text):
		text = text.strip ()
		return [a for a in cls.by_tag (node, "a") if a.get_text () == text]			
	
	@classmethod
	def by_plt (cls, node, text):
		text = text.strip ()
		return [a for a in cls.by_tag (node, "a") if a.get_text ().find (text) != -1]
	
	@classmethod
	def by_css_data (cls, node, text):
		css, text = text.split (":data", 1)
		css = css.strip ()
		text = text.strip ()
		element = cls.by_css (node, css)
		return filter_by_data (element, text)
		
	@classmethod
	def by_tint (cls, node, text):
		tag, text = text.split ("=", 1)
		text = "=" + text.strip ()
		tag = tag.strip ()
		if not tag [-1].isalpha ():
			text = tag [-1] + text
			tag = tag [:-1]
		text = text.strip ()
		element = cls.by_tag (node, tag)
		return filter_by_data (element, text)
								
	@classmethod
	def prev_siblings (cls, node):
		nodes = []
		for c, n in enumerate (node.itersiblings(preceding=True)):
			nodes.append(n)
		return nodes
		
	@classmethod
	def prev_sibling (cls, node):
		nodes = []
		for c, n in enumerate (node.itersiblings(preceding=True)):
			nodes.append(n)
			if c == 0:
				break
		return nodes[0] if nodes else None
	
	@classmethod
	def next_siblings (cls, node):
		nodes = []
		for c, n in enumerate (node.itersiblings(preceding=False)):
			nodes.append(n)
		return nodes
		
	@classmethod
	def next_sibling (cls, node):
		nodes = []
		for c, n in enumerate (node.itersiblings(preceding=False)):
			nodes.append(n)
			if c == 0:
				break
		return nodes[0] if nodes else None
	
	@classmethod
	def get_siblings (cls, node):
		return cls.prev_siblings (node) + cls.next_siblings (node)
		
	@classmethod
	def new (self, tag):			
		return lxml.etree.Element(tag)
		
	@classmethod
	def get_attr (cls, node, attr=None, default = None):
		if attr:
			val = node.attrib.get(attr, None)
			if not val:
				return default
			return val	
		return node.attrib
	
	@classmethod
	def has_attr (cls, node, attr = None):		
		if attr:
			return attr in node.attrib
		return len (node.attrib) != 0
		
	@classmethod
	def del_attr (cls, node, attr=None):
		if attr:
			_attr = node.attrib.get(attr, None)
			if _attr:
				del node.attrib[attr]

	@classmethod
	def set_attr (cls, node, attr=None, value=None):
		if attr and value:
			node.set(attr, value)
					
	@classmethod
	def append_child (cls, node, child):
		node.append(child)
	
	@classmethod
	def get_path (cls, node):
		tree = lxml.etree.ElementTree(node)
		return tree.getpath(node)
		
	@classmethod
	def insert_child (cls, index, node, child):
		node.insert (index, child)
	
	@classmethod
	def insert_next (cls, node, child):
		index = 0
		p = cls.get_parent (node)
		assert p, "No prarent"
		for sib in cls.get_children (p):
			if sib == node:				
				break
			index += 1		
		p.insert (index + 1, child)
	
	@classmethod
	def insert_before (cls, node, child):
		index = 0
		p = cls.get_parent (node)		
		assert p, "No prarent"
		for sib in cls.get_children (p):
			if sib == node:				
				break
			index += 1
		p.insert (index, child)
			
	@classmethod
	def child_nodes (cls, node):
		return list(node)

	@classmethod
	def child_nodes_with_text (cls, node):
		root = node
		# create the first text node
		# if we have some text in the node
		if root.text:
			t = lxml.html.HtmlElement()
			t.text = root.text
			t.tag = 'text'
			root.text = None
			root.insert(0, t)
		# loop childs
		for c, n in enumerate(list(root)):
			idx = root.index(n)
			# don't process texts nodes
			if n.tag == 'text':
				continue
			# create a text node for tail
			if n.tail:
				t = cls.create_element(tag='text', text=n.tail, tail=None)
				root.insert(idx + 1, t)
		return list(root)

	@classmethod
	def get_children (cls, node):
		return node.getchildren()
	
	@classmethod
	def get_parent (cls, node):
		return node.getparent()
		
	@classmethod
	def get_text (cls, node):
		txts = cls.get_text_list (node, False)
		return inner_trim (' '.join(txts).strip())
	
	@classmethod
	def get_texts (cls, node, trim = True):
		return cls.get_text_list (node, trim)
	
	@classmethod
	def get_text_list (cls, node, trim = True):
		txts = list (node.itertext())
		#if node.tail and node.tail.strip ():
		#	txts.append (node.tail)
		if trim:
			return [i.strip () for i in txts]
		return txts
	
	@classmethod
	def has_tail (self, node, recursive = 0):
		if node.tail and node.tail.strip ():
			return 1				
		if recursive:	
			for child in nops.get_children (node):
				has_tail = self.has_tail (child, 1)
				if has_tail:
					return 1
				
	@classmethod
	def collect_children (cls, node):
		def collect (node, container):
			children = cls.get_children (node)
			if not children:
				return
				
			for child in children:								
				container.append (child)
				collect (child, container)

		container = []
		collect (node, container)
		return container
		
	@classmethod
	def iter_text (cls, node):
		def collect (node, container):
			children = cls.get_children (node)
			if not children:
				return
				
			for child in children:
				if child.tag not in (
					"head", "meta", "link", "input", "hr", "br", "img", 
					"table", "tr", "thead", "tbody", "ol", "ul", "dl"
				):
					text = child.text
					if text is not None:
						text = text.strip ()
					else:
						text = ""
					container.append (text)
				collect (child, container)

		container = []
		collect (node, container)
		return container

	@classmethod
	def is_text_node (cls, node):
		return True if node.tag == 'text' else False
	
	@classmethod
	def get_tag (cls, node):
		return node.tag
	
	@classmethod
	def replace_tag (cls, node, tag):
		node.tag = tag

	@classmethod
	def strip_tags (cls, node, *tags):
		lxml.etree.strip_tags(node, *tags)
			
	@classmethod
	def drop_node (cls, node):		
		try: 
			node.drop_tag ()
		except AttributeError:			
			p = node.getparent ()
			for child in node.getchildren ():
				cls.insert_before (node, child)							
			p.remove (node)
	
	@classmethod
	def drop_tree (cls, node):
		def recursive (node):
			for child in node.getchildren ():
				if child.getchildren ():
					recursive (child)
				else:
					node.remove (child)

		try: 
			node.drop_tree ()
		except AttributeError:
			recursive (node)
			node.getparent ().remove (node)
	
	@classmethod
	def contains (cls, node, doc):
		def find_recursive (node, doc):
			children = doc.getchildren ()
			if node in children:
				return True
			for child in children:
				is_contain = find_recursive (node, child)
				if is_contain:
					return True
			return False							
		return find_recursive (node, doc)
			
	@classmethod
	def create_element (cls, tag='p', text=None, tail=None):
		t = lxml.html.HtmlElement()
		t.tag = tag
		t.text = text
		t.tail = tail
		return t

	@classmethod
	def get_comments (cls, node):
		return node.xpath('//comment()')
	
	@classmethod
	def remove_comments (cls, node):
		for item in cls.get_comments (node):
			try:
				item.drop_tree ()
			except AssertionError: # out of root node
				pass
				
	@classmethod
	def text_to_para (cls, text):
		return cls.create_element ('p', text)
	
	@classmethod
	def outer_html (cls, node):
		e0 = node
		if e0.tail:
			e0 = deepcopy(e0)
			e0.tail = None
		return cls.to_string(e0)
	
	ALLOW_TAGS = [
		'a', 'span', 'p', 'br', 'strong', 'b',
		'em', 'i', 'tt', 'code', 'pre', 'blockquote', 'img', 'h1',
		'h2', 'h3', 'h4', 'h5', 'h6'
	]
		
	@classmethod
	def clean_html (cls, node, allow_tags = None):
		article_cleaner = lxml.html.clean.Cleaner()
		article_cleaner.javascript = True
		article_cleaner.style = True
		if not allow_tags:
			article_cleaner.allow_tags = cls.ALLOW_TAGS
		else:
			article_cleaner.allow_tags = allow_tags			
		article_cleaner.remove_unknown_tags = False
		return article_cleaner.clean_html (node)
	
	@classmethod
	def get_param (cls, node, attr, name):
		name = name.lower ()
		params = cls.get_attr (node, attr)
		for param in params.split (";"):
			param = param.strip ()
			if not param.lower ().startswith (name):
				continue
			
			val = param [len (name):].strip ()
			if not val: return ""
			if val [0] == "=":
				val = val [1:].strip ()
				if not val: return ""
			if val [0] in "\"'":
				return val [1:-1]
			return val		
	
	@classmethod
	def resolve_links (cls, node, base_href = None):
		if base_href:
			try:
				# such support is added in lxml 3.3.0
				node.make_links_absolute (base_href, resolve_base_href=True, handle_failures='discard')
			except TypeError:
				node.make_links_absolute (base_href, resolve_base_href=True)
		else:
			node.resolve_base_href ()
		return node
	
	@classmethod
	def get_child_has_grand_children_or_none (cls, origin_node):
		node = origin_node
		while 1:
			children = cls.get_children (node)
			if not children or len (children) > 1:
				return node
			node = children [0]
	
	@classmethod
	def add_class (cls, node, name):
		classes = cls.get_attr (node, "class", "").split (" ")		
		for c in name.split (" "):
			if c not in classes:
				classes.append (c)
		cls.set_attr (node, "class", " ".join (classes).strip ())
	
	@classmethod
	def remove_class (cls, node, name):
		classes = [c for c in cls.get_attr (node, "class", "").split (" ") if c != name]
		new_class = " ".join (classes).strip ()
		if not new_class:
			cls.del_attr (node, "class")
		else:	
			cls.set_attr (node, "class", new_class)		
	
	@classmethod	
	def has_class (cls, node, name):
		return [c for c in cls.get_attr (node, "class", "").split (" ") if c == name]
	
	@classmethod	
	def update_class (cls, node, name, new_name):
		cls.remove_class (node, name)
		cls.add_class (node, new_name)		
	
	@classmethod
	def batch (cls, node, batch):
		for cmd, param in batch:
			if cmd == "strip":
				cls.strip_tags (node, param)				
			
			elif cmd in ("replace", "wrap", "insertb", "inserta", "attach"):
				t = param.split (" ")
				a = " ".join (t[:-1]).strip ()
				b = t[-1].strip ()
				for each in cls.by_css (node, a):
					if cmd == "replace":
						each.tag = b
					elif cmd == "attach":
						n = cls.new (b)
						cls.append_child (each, n)
					elif cmd == "inserta":
						n = cls.new (b)
						cls.insert_after (each, n)
					elif cmd == "insertb":
						n = cls.new (b)
						cls.insert_before (each, n)	
					else:
						n = cls.new (b)
						cls.insert_next (each, n)
						cls.append_child (n, each)
							
			else:
				method, target =  cssutil.make_detector (*param)
				elts = method (node, target)				
				if cmd == "select":
					body = nops.new ("body")
					for elt in elts:
						cls.append_child (body, elt)
					for each in cls.by_tag (node, "body"):
						cls.drop_tree (each)
					cls.append_child (node, body)
					
				else:
					for elt in elts:
						if cmd [:6] == "class+":							
							cls.add_class (elt, cmd [6:])
						elif cmd [:6] == "class-":
							cls.remove_class (elt, cmd [6:])
						elif cmd in ("deltree", "rmtree"):
							cls.drop_tree (elt)
						elif cmd in ("unwrap", "delnode", "rmnode"):
							cls.drop_node (elt)
						elif cmd == "rmtail":
							elt.tail = None
						elif cmd in ("clear", "rmtext"):
							elt.text = ""
						elif cmd == "backspace":
							elt.text += " "
						elif cmd == "frontspace":
							elt.text = " " + elt.text
						else:
							raise SystemError ('Unknown command: %s' % cmd)
						
	@classmethod
	def walk_siblings (cls, node):
		current_sibling = cls.prev_sibling (node)
		b = []
		while current_sibling is not None:
			b.append (current_sibling)
			current_sibling = cls.prev_sibling (current_sibling)
		return b
	
	@classmethod
	def is_decendent (cls, target, node):
		if target == node:
			return 0
		for child in cls.get_children (node):
			if child == target:
				return 1
			if cls.is_decendent (target, child):
				return 1
		return 0	
	
	@classmethod
	def is_ancestor (cls, target, node):
		if target == node:
			return 0
		p = node
		while p:			
			if p == target:
				return 1
			p = cls.get_parent (p)
		return 0	
		
	@classmethod
	def sort_decendent (cls, ancestor, nodes, sorted_children = None):
		if sorted_children is None:
			sorted_children = []
		for child in cls.get_children (ancestor):
			if child in nodes and child not in sorted_children:
				sorted_children.insert (0, child)
				if len (nodes) == len (sorted_children):
					return sorted_children
			cls.sort_decendent (child, nodes, sorted_children)
		return sorted_children	
															
if __name__ == "__main__":	
	from urllib.request import urlopen
	from contextlib import closing
	
	with closing(urlopen("http://www.drugandalcoholrehabhouston.com")) as f:
		build (f.read ())



import time
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select
from selenium.common.exceptions import TimeoutException
from .nops import nops
import os

DEFAULT_USER_AGENT = "Mozilla/5.0 (compatible; rfphound/1.0; +http://rfphound.com/bot; AppleWebKit/537.36 (KHTML, like Gecko) Chrome/66.0.3359.181 Safari/537.36)"


OPS = {
        "*": "contains",
        "^": "starts-with",
        "$": "ends-with"
}
def _transport (what, by):
    if what.startswith ("link-text:"):
        what = "a:data*|" + what [10:]
    elif what.startswith ("button-text:"):
        what = "button:data*|" + what [12:]

    s = what.find (':data')
    if s == -1:
        return what, By.CSS_SELECTOR

    tag, text = what [:s], what [s + 5:]
    by = By.XPATH

    op = "="
    include_descendant = False
    if text [1] not in "=|":
        assert text [0] == "=", "Unkown Operator {}".format (text [0])
        text = text [1:]
    else:
        if text [1] == "|":
            include_descendant = True
        op = text [0]
        assert op in OPS, "Unkown Operator {}".format (text [:2])
        text = text [2:]

    if op == "=":
        if include_descendant:
            what = "//{}[.//*[text() = '{}']]".format (tag, text.replace ("'", "\\'"))
        else:
            what = "//{}[text() = '{}']".format (tag, text.replace ("'", "\\'"))

    else:
        if include_descendant:
            what = "//{}[.//*[{}(text(), '{}')]]".format (tag, OPS [op], text.replace ("'", "\\'"))
        else:
            what = "//{}[{}(text(), '{}')]".format (tag, OPS [op], text.replace ("'", "\\'"))
    return what, by


class KeysProxy:
    def __getattr__ (self, key):
        # BACKSPACE DELETE RETURN HOME END etc
        return getattr (Keys, key)


class Driver:
    def __init__ (self, driver_path, headless = False, user_agent = DEFAULT_USER_AGENT, window_size = (1600, 900), **kargs):
        self.driver_path = driver_path
        self.capabilities = self.setopt (headless, user_agent, window_size, **kargs)
        self.driver = None
        self.keys = KeysProxy ()
        self.main_window = None

    # -----------------------------------------------------------------
    def _parse_host (self, host):
        try:
            server, port = host.split (":", 1)
        except ValueError:
            raise ValueError ("proxy port is required")
        port = int (port)
        return server, port

    def __getattr__ (self, attr):
        if self.driver:
            return getattr (self.driver, attr)
        raise AttributeError ("driver does not be stared or driver has not '{}' attribute".format (attr))

    def __enter__ (self):
        if self.driver is None:
            self.create ()
        return self

    def __exit__ (self, type, value, tb):
        self.driver.quit ()
        self.driver = None

    # callables ---------------------------------------------------------
    def set_timeout (self, timeout):
        self.driver.set_page_load_timeout (timeout)

    def implicitly_wait (self, timeout):
        self.driver.implicitly_wait (timeout)

    def get (self, url):
        self.main_window = self.driver.current_window_handle
        self.driver.get (url)
    navigate = get

    def new_tab (self, url, window_name):
        self.driver.execute_script (
            "(function () {"
            'window.open ("%s", "%s")'
            "}) ()" % (url, window_name)
        )

    def html (self, timeout = 5):
        self.wait (timeout)
        return "<html>%s</html>" % self.driver.execute_script ("return document.getElementsByTagName('html')[0].innerHTML;").replace ("&nbsp;", " ")
    get_html = html

    def lxml (self):
        return nops.from_string (self.html ())

    def sleep (self, timeout):
        time.sleep (timeout)

    def _wait (self, timeout = 5, until = 'loaded', by = By.CSS_SELECTOR, what = None):
        def document_loaded (driver):
            return driver.execute_script ("return document.readyState;") == "complete"

        if until == "loaded":
            return WebDriverWait (self.driver, timeout).until (document_loaded)

        cond = {
            'title-like': EC.title_contains,
            'title': EC.title_is,
            'url-like': EC.url_contains,
            'url': EC.url_to_be,
            '~url': EC.url_matches,
        }.get (until)
        if cond:
            return WebDriverWait (self.driver, timeout).until (cond (what))

        cond = {
            'presence': EC.presence_of_element_located,
            'clickable': EC.element_to_be_clickable,
            'visible': EC.visibility_of_element_located,
            'invisible': EC.invisibility_of_element_located,

        }.get (until)
        if cond:
            return WebDriverWait (self.driver, timeout).until (cond ((by, what)))

        raise ValueError ("Unknown Condition")

    # htmlops compatible---------------------------------------
    def wait (self, timeout = 5):
        return self._wait (timeout, 'loaded')

    def query (self, by, what, timeout = 5, until = 'presence'):
        what, by = _transport (what, by)
        if timeout:
            self._wait (timeout, until, by, what)
        return self.driver.find_elements (by, what)

    # for e2e test ---------------------------------------
    def capture (self, path = './screenshot.jpg'):
        self.save_screenshot (path)

    def wait_until (self, until = 'presence', what = None, timeout = 5, by = By.CSS_SELECTOR):
        return self._wait (timeout, until, by, what)

    def fetch (self, what, until = 'presence', timeout = 5, by = By.CSS_SELECTOR):
        try:
            return self.query (by, what, timeout, until)
        except TimeoutException:
            return

    def one (self, what, until = 'presence', timeout = 5, by = By.CSS_SELECTOR):
        try:
            return self.fetch (what, until, timeout, by) [0]
        except (TypeError, IndexError):
            return None

    def as_select (self, what, until = 'presence', timeout = 5, by = By.CSS_SELECTOR):
        item = self.one (what, until, timeout, by)
        if item:
            return Select (item)

    def run (self, script):
        return self.driver.execute_script (script)

    # inputs handling --------------------------------------------------
    # SELECT
    def select (self, s, index):
        s = Select (s)
        if type (index) is int:
            s.select_by_index(index)
        else:
            s.select_by_value(index)

    def unselect (self, s, index):
        s = Select (s)
        if type (index) is int:
            s.deselect_by_index(int (index))
        else:
            s.deselect_by_value(index)

    def deselect_all (self, s):
        s = Select (s)
        s.deselect_all ()

    def select_all (self, s, text):
        s = Select (s)
        s.select_all ()

    def select_by_text (self, s, text):
        s = Select (s)
        s.select_by_visible_text(text)

    def unselect_by_text (self, s, text):
        s = Select (s)
        s.deselect_by_visible_text(text)

    # CHECKBOX, RADIO
    def check (self, group, thing):
        if type (thing) is int:
            if not group [thing].is_selected():
                group [thing].click ()
            return

        for cb in group:
            if cb.get_attribute ("value") == thing and not cb.is_selected():
                cb.click()

    def uncheck (self, group, thing):
        if type (thing) is int:
            if group [thing].is_selected():
                group [thing].click ()
            return

        for cb in group:
            if cb.get_attribute ("value") == thing and cb.is_selected():
                cb.click()

    def uncheck_all (self, group):
        for cb in group:
            if cb.is_selected():
                cb.click ()

    def check_all (self, group):
        for cb in group:
            if not cb.is_selected():
                cb.click ()

    # TEXT, PASSWORD, TEXTAREA
    def set_text (self, e, *text):
        e.send_keys (*text)
    send_keys = set_text

    def save_html (self, path):
        with open (path, "w") as f:
            f.write (self.html ())

    # shortcuts ---------------------------------------------------
    def get_frames (self, for_human = True, timeout = 10):
        self.wait (timeout)
        names = []
        index = 0
        for ftype in ("frame", "iframe"):
            frames = self.driver.find_elements_by_tag_name(ftype)
            for frame in frames:
                if for_human:
                    names.append (
                        "FRAME #%d. type: %s, name: %s, width: %s, height: %s, src: %s" % (
                        index, ftype,
                        frame.get_attribute ("name"),
                        frame.get_attribute ("width"), frame.get_attribute ("height"),
                        frame.get_attribute ("src")
                        )
                    )
                else:
                    names.append (frame)
                index += 1
        return names

    def get_windows (self, for_human = True):
        windows = []
        index = 0
        for w in self.driver.window_handles:
            if w == self.main_window:
                continue

            if for_human:
                self.driver.switch_to.window (w)
                windows.append ("WINDOW #%d. name: %s, url: %s, title: %s" % (
                        index,
                        self.driver.name,
                        self.driver.current_url,
                        self.driver.title
                    )
                )
            else:
                windows.append (w)
            index += 1
        #self.driver.switch_to.window (self.main_window)
        return windows

    def switch_to_window (self, index = None):
        if index is None:
            return self.driver.switch_to.window (self.main_window)
        if isinstance (index, str):
            return self.driver.switch_to.window (index)
        return self.driver.switch_to.window (self.get_windows (False)[index])

    def switch_to_frame (self, index = None):
        if index is None:
            return self.driver.switch_to.default_content()
        if isinstance (index, str):
            return self.driver.switch_to.frame (index)
        return self.driver.switch_to.frame (self.get_frames (False)[index])

    def switch_to_active (self):
        return self.driver.switch_to.active_element

    def switch_to_alert (self):
        return self.driver.switch_to.alert()

    def close_window (self, switch_to = None):
        self.driver.close ()
        return self.switch_to_window (switch_to)
    close = close_window

    # @abstractmethod ----------------------------------------------
    def setopt (self, headless, user_agent, window_size, **opts):
        pass

    def create (self):
        pass


import zlib
import gzip
import io
import time
import struct
import sys

PY_MAJOR_VERSION = sys.version_info.major

class DeflateCompressor:
	def __init__ (self, level = 5):
		self.compressor = zlib.compressobj (5, zlib.DEFLATED)
		
	def compress (self, buf):	
		return self.compressor.compress (buf)
	
	def flush (self):
		return self.compressor.flush ()


class DeflateDeompressor:
	def __init__ (self):
		self.decompressor = zlib.decompressobj ()	
		
	def compress (self, buf):	
		d = self.decompressor.decompress (buf)
		if d == b"":
			return self.flush ()
		return d	
	
	def flush (self):
		return self.decompressor.flush ()
		
				
class GZipCompressor:
	#HEADER = b"\037\213\010" + chr (0) + struct.pack ("<L", int (time.time ())) + b"\002\377"
	HEADER = b"\037\213\010" + b'\x00' + struct.pack ("<L", int (time.time ())) + b"\002\377"	
	def __init__ (self, level = 5):
		self.size = 0
		self.crc = zlib.crc32(b"")
		self.compressor = zlib.compressobj (level, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0)
		self.first_data = True
			
	def compress (self, buf):
		self.size = self.size + len(buf)
		self.crc = zlib.crc32(buf, self.crc)
		d = self.compressor.compress (buf)
		if self.first_data:
			d = self.HEADER + d
			self.first_data = False
		return d
	
	def flush (self):
		d = self.compressor.flush ()
		if PY_MAJOR_VERSION	>= 3:
			return d + struct.pack ("<L", self.crc) + struct.pack ("<L", self.size & 0xFFFFFFFF)
		else:
			return d + struct.pack ("<l", self.crc) + struct.pack ("<L", self.size & 0xFFFFFFFF)	
		

def U32(i):
	if i < 0:
		i += 1 << 32
	return i
    
class GZipDecompressor:	
	def __init__ (self, level = 5):
		self.size = 0
		self.crc = zlib.crc32(b"")
		self.decompressor = zlib.decompressobj(-zlib.MAX_WBITS)
		self.first_data = True
		self.maybe_eof = b""
		self.extra = False
			
	def decompress (self, buf):
		if self.first_data:
			buf = buf [10:]
			self.first_data = False
		elif self.extra:
			if len (buf) == 8:
				return ""
			else:
				self.extra = False
				buf = b"\x03\x00" + buf
		
		if buf == b"\x03\x00":
			# what is it?	
			self.extra = True
				
		if len (buf) > 8:
			self.maybe_eof = buf [-8:]
		else:
			self.maybe_eof += buf
			self.maybe_eof = self.maybe_eof [-8:]
		
		d = self.decompressor.decompress (buf)
		self.size += len (d)
		self.crc = zlib.crc32(d, self.crc)
		if d == b"":			
			return self.decompressor.flush ()
		return d
	
	def flush (self):
		crcs, isizes = self.maybe_eof [:4], self.maybe_eof [4:]
		if PY_MAJOR_VERSION	>= 3:
			crc32 = struct.unpack ("<L", crcs)[0]
		else:
			crc32 = struct.unpack ("<l", crcs)[0]	
		isize = U32 (struct.unpack ("<L", isizes)[0])
		if U32 (crc32) != U32 (self.crc):
			raise IOError("CRC check failed")
		elif isize != (self.size & 0xFFFFFFFF):
			raise IOError("Incorrect length of data produced")
		return b""
			

if __name__ == "__main__":
	import urllib.request, urllib.parse, urllib.error
	f =urllib.request.urlopen ("http://www.gmarket.co.kr/index.asp/")
	d = f.read ()
	
	a = GZipCompressor ()
	x = a.compress (d) + a.flush ()	
	b = GZipDecompressor ()	
	while x:
		k, x = x [:10], x [10:]		
		b.decompress (k)
	print(repr(b.flush ()))
	print(repr(b.flush ()))
	print(repr(b.flush ()))	
	a = zlib.compressobj ()
	x = a.compress (d) + a.flush ()	
	b = zlib.decompressobj ()	
	while x:
		k, x = x [:10], x [10:]				
		b.decompress (k)
	b.decompress ("")	
	b.decompress ("")	
	print(repr(b.flush ()))
	print(repr(b.flush ()))
	print(repr(b.flush ()))
	



# Copyright (C) 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for audio streams."""

import array
import logging
import math
import time
import threading
import wave

import click
import sounddevice as sd


DEFAULT_AUDIO_SAMPLE_RATE = 16000
DEFAULT_AUDIO_SAMPLE_WIDTH = 2
DEFAULT_AUDIO_ITER_SIZE = 3200
DEFAULT_AUDIO_DEVICE_BLOCK_SIZE = 6400
DEFAULT_AUDIO_DEVICE_FLUSH_SIZE = 25600


def normalize_audio_buffer(buf, volume_percentage, sample_width=2):
    """Adjusts the loudness of the audio data in the given buffer.

    Volume normalization is done by scaling the amplitude of the audio
    in the buffer by a scale factor of 2^(volume_percentage/100)-1.
    For example, 50% volume scales the amplitude by a factor of 0.414,
    and 75% volume scales the amplitude by a factor of 0.681.
    For now we only sample_width 2.

    Args:
      buf: byte string containing audio data to normalize.
      volume_percentage: volume setting as an integer percentage (1-100).
      sample_width: size of a single sample in bytes.
    """
    if sample_width != 2:
        raise Exception('unsupported sample width:', sample_width)
    scale = math.pow(2, 1.0*volume_percentage/100)-1
    # Construct array from bytes based on sample_width, multiply by scale
    # and convert it back to bytes
    arr = array.array('h', buf)
    for idx in range(0, len(arr)):
        arr[idx] = int(arr[idx]*scale)
    buf = arr.tostring()
    return buf


def align_buf(buf, sample_width):
    """In case of buffer size not aligned to sample_width pad it with 0s"""
    remainder = len(buf) % sample_width
    if remainder != 0:
        buf += b'\0' * (sample_width - remainder)
    return buf


class WaveSource(object):
    """Audio source that reads audio data from a WAV file.

    Reads are throttled to emulate the given sample rate and silence
    is returned when the end of the file is reached.

    Args:
      fp: file-like stream object to read from.
      sample_rate: sample rate in hertz.
      sample_width: size of a single sample in bytes.
    """
    def __init__(self, fp, sample_rate, sample_width):
        self._fp = fp
        try:
            self._wavep = wave.open(self._fp, 'r')
        except wave.Error as e:
            logging.warning('error opening WAV file: %s, '
                            'falling back to RAW format', e)
            self._fp.seek(0)
            self._wavep = None
        self._sample_rate = sample_rate
        self._sample_width = sample_width
        self._sleep_until = 0

    def read(self, size):
        """Read bytes from the stream and block until sample rate is achieved.

        Args:
          size: number of bytes to read from the stream.
        """
        now = time.time()
        missing_dt = self._sleep_until - now
        if missing_dt > 0:
            time.sleep(missing_dt)
        self._sleep_until = time.time() + self._sleep_time(size)
        data = (self._wavep.readframes(size)
                if self._wavep
                else self._fp.read(size))
        #  When reach end of audio stream, pad remainder with silence (zeros).
        if not data:
            return b'\x00' * size
        return data

    def close(self):
        """Close the underlying stream."""
        if self._wavep:
            self._wavep.close()
        self._fp.close()

    def _sleep_time(self, size):
        sample_count = size / float(self._sample_width)
        sample_rate_dt = sample_count / float(self._sample_rate)
        return sample_rate_dt

    def start(self):
        pass

    def stop(self):
        pass

    @property
    def sample_rate(self):
        return self._sample_rate


class WaveSink(object):
    """Audio sink that writes audio data to a WAV file.

    Args:
      fp: file-like stream object to write data to.
      sample_rate: sample rate in hertz.
      sample_width: size of a single sample in bytes.
    """
    def __init__(self, fp, sample_rate, sample_width):
        self._fp = fp
        self._wavep = wave.open(self._fp, 'wb')
        self._wavep.setsampwidth(sample_width)
        self._wavep.setnchannels(1)
        self._wavep.setframerate(sample_rate)

    def write(self, data):
        """Write bytes to the stream.

        Args:
          data: frame data to write.
        """
        self._wavep.writeframes(data)

    def close(self):
        """Close the underlying stream."""
        self._wavep.close()
        self._fp.close()

    def start(self):
        pass

    def stop(self):
        pass

    def flush(self):
        pass


class SoundDeviceStream(object):
    """Audio stream based on an underlying sound device.

    It can be used as an audio source (read) and a audio sink (write).

    Args:
      sample_rate: sample rate in hertz.
      sample_width: size of a single sample in bytes.
      block_size: size in bytes of each read and write operation.
      flush_size: size in bytes of silence data written during flush operation.
    """
    def __init__(self, sample_rate, sample_width, block_size, flush_size, device = None):
        if sample_width == 2:
            audio_format = 'int16'
        else:
            raise Exception('unsupported sample width:', sample_width)
        self._audio_stream = sd.RawStream(
            samplerate=sample_rate, dtype=audio_format, channels=1, device = device,
            blocksize=int(block_size/2),  # blocksize is in number of frames.
        )
        self._block_size = block_size
        self._flush_size = flush_size
        self._sample_rate = sample_rate

    def read(self, size):
        """Read bytes from the stream."""
        buf, overflow = self._audio_stream.read(size)
        if overflow:
            logging.warning('SoundDeviceStream read overflow (%d, %d)',
                            size, len(buf))
        return bytes(buf)

    def write(self, buf):
        """Write bytes to the stream."""
        underflow = self._audio_stream.write(buf)
        if underflow:
            logging.warning('SoundDeviceStream write underflow (size: %d)',
                            len(buf))
        return len(buf)

    def flush(self):
        if self._audio_stream.active and self._flush_size > 0:
            self._audio_stream.write(b'\x00' * self._flush_size)

    def start(self):
        """Start the underlying stream."""
        if not self._audio_stream.active:
            self._audio_stream.start()

    def stop(self):
        """Stop the underlying stream."""
        if self._audio_stream.active:
            self._audio_stream.stop()

    def close(self):
        """Close the underlying stream and audio interface."""
        if self._audio_stream:
            self.stop()
            self._audio_stream.close()
            self._audio_stream = None

    @property
    def sample_rate(self):
        return self._sample_rate


class ConversationStream(object):
    """Audio stream that supports half-duplex conversation.

    A conversation is the alternance of:
    - a recording operation
    - a playback operation

    Excepted usage:

      For each conversation:
      - start_recording()
      - read() or iter()
      - stop_recording()
      - start_playback()
      - write()
      - stop_playback()

      When conversations are finished:
      - close()

    Args:
      source: file-like stream object to read input audio bytes from.
      sink: file-like stream object to write output audio bytes to.
      iter_size: read size in bytes for each iteration.
      sample_width: size of a single sample in bytes.
    """
    def __init__(self, source, sink, iter_size, sample_width):
        self._source = source
        self._sink = sink
        self._iter_size = iter_size
        self._sample_width = sample_width
        self._volume_percentage = 50
        self._stop_recording = threading.Event()
        self._source_lock = threading.RLock()
        self._recording = False
        self._playing = False

    def start_recording(self):
        """Start recording from the audio source."""
        self._recording = True
        self._stop_recording.clear()
        self._source.start()

    def stop_recording(self):
        """Stop recording from the audio source."""
        self._stop_recording.set()
        with self._source_lock:
            self._source.stop()
        self._recording = False

    def start_playback(self):
        """Start playback to the audio sink."""
        self._playing = True
        self._sink.start()

    def stop_playback(self):
        """Stop playback from the audio sink."""
        self._sink.flush()
        self._sink.stop()
        self._playing = False

    @property
    def recording(self):
        return self._recording

    @property
    def playing(self):
        return self._playing

    @property
    def volume_percentage(self):
        """The current volume setting as an integer percentage (1-100)."""
        return self._volume_percentage

    @volume_percentage.setter
    def volume_percentage(self, new_volume_percentage):
        self._volume_percentage = new_volume_percentage

    def read(self, size):
        """Read bytes from the source (if currently recording).
        """
        with self._source_lock:
            return self._source.read(size)

    def write(self, buf):
        """Write bytes to the sink (if currently playing).
        """
        buf = align_buf(buf, self._sample_width)
        buf = normalize_audio_buffer(buf, self.volume_percentage)
        return self._sink.write(buf)

    def close(self):
        """Close source and sink."""
        self._source.close()
        self._sink.close()

    def __iter__(self):
        """Returns a generator reading data from the stream."""
        while True:
            if self._stop_recording.is_set():
                return
            yield self.read(self._iter_size)

    @property
    def sample_rate(self):
        return self._source._sample_rate


@click.command()
@click.option('--record-time', default=5,
              metavar='<record time>', show_default=True,
              help='Record time in secs')
@click.option('--audio-sample-rate',
              default=DEFAULT_AUDIO_SAMPLE_RATE,
              metavar='<audio sample rate>', show_default=True,
              help='Audio sample rate in hertz.')
@click.option('--audio-sample-width',
              default=DEFAULT_AUDIO_SAMPLE_WIDTH,
              metavar='<audio sample width>', show_default=True,
              help='Audio sample width in bytes.')
@click.option('--audio-iter-size',
              default=DEFAULT_AUDIO_ITER_SIZE,
              metavar='<audio iter size>', show_default=True,
              help='Size of each read during audio stream iteration in bytes.')
@click.option('--audio-block-size',
              default=DEFAULT_AUDIO_DEVICE_BLOCK_SIZE,
              metavar='<audio block size>', show_default=True,
              help=('Block size in bytes for each audio device '
                    'read and write operation..'))
@click.option('--audio-flush-size',
              default=DEFAULT_AUDIO_DEVICE_FLUSH_SIZE,
              metavar='<audio flush size>', show_default=True,
              help=('Size of silence data in bytes written '
                    'during flush operation'))
def main(record_time, audio_sample_rate, audio_sample_width,
         audio_iter_size, audio_block_size, audio_flush_size):
    """Helper command to test audio stream processing.

    - Record 5 seconds of 16-bit samples at 16khz.
    - Playback the recorded samples.
    """
    end_time = time.time() + record_time
    audio_device = SoundDeviceStream(sample_rate=audio_sample_rate,
                                     sample_width=audio_sample_width,
                                     block_size=audio_block_size,
                                     flush_size=audio_flush_size)
    stream = ConversationStream(source=audio_device,
                                sink=audio_device,
                                iter_size=audio_iter_size,
                                sample_width=audio_sample_width)
    samples = []
    logging.basicConfig(level=logging.INFO)
    logging.info('Starting audio test.')

    stream.start_recording()
    logging.info('Recording samples.')
    while time.time() < end_time:
        samples.append(stream.read(audio_block_size))
    logging.info('Finished recording.')
    stream.stop_recording()

    stream.start_playback()
    logging.info('Playing back samples.')
    while len(samples):
        stream.write(samples.pop(0))
    logging.info('Finished playback.')
    stream.stop_playback()

    logging.info('audio test completed.')
    stream.close()


if __name__ == '__main__':
    main()



import urllib.request, urllib.parse, urllib.error
import re
import time
from ...client.http import ClientCookie
from ...lib import timeoutsocket, confparse
from ...lib import logger as loggerfactory
import tempfile
import os
import random
from ...lib import pathtool
from . import __init__
import math

n, p = pathtool.modpath (__init__)
PROXYURLS = "proxysite.dat"

rx_ip = r"[^0-9]([0-9]{2,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})"
rx_dns = r"[^a-z0-9]([0-9a-z][-a-z0-9_]{2,30}\.[a-z0-9_]{2,30}\.[a-z0-9_.]*?)"
rx_port = r"([0-9]{2,5})"

RX_PROXY = [
		re.compile ("%s:%s" % (rx_ip, rx_port), re.I),
		re.compile ("%s.+?:%s" % (rx_ip, rx_port)),
		re.compile ("%s</td>\s*?<td.*?>%s</td>" % (rx_ip, rx_port), re.M),	
		re.compile ("%s\s+port\s+%s</" % (rx_ip, rx_port)),	
		re.compile ("%s:%s" % (rx_dns, rx_port), re.I),
		re.compile ("%s</td>\s*?<td.*?>%s</td>" % (rx_dns, rx_port), re.I|re.M),
		re.compile ("%s\s+port\s+%s</" % (rx_dns, rx_port)),		
		re.compile ("%s.+?>\s*%s" % (rx_ip, rx_port), re.I|re.S),
]

def unhide_text (data):
	t=''
	s = urllib.parse.unquote (data)
	x = int (round(math.sqrt(49)))
	for c in s: 		
 		t += chr (ord(c)^x)
 	t = t.replace ("#", " ").replace ("!", " ").replace ("|", " ").replace ("-", "")
 	return t

def unhide_text2 (data):
	t=''
	s = urllib.parse.unquote (data)
	for c in s: 		
 		t += chr (ord(c)^4) 	
 	return t
 	

class ProxyFinder:
	def __init__ (self, logger = None):
		self.logger = logger
		self.failover = 1
		self.fn = os.path.join (tempfile.gettempdir (), "__proxy.dat")
		if not self.logger:
			self.logger = loggerfactory.screen_logger ()
		self.proxies = {}
		self.hits = 0
		self.hold = False
		self.cookies = ClientCookie.CookieJar()
		self.opener = ClientCookie.build_opener(ClientCookie.HTTPCookieProcessor(self.cookies))
		self.goods = {}
	
	def set_hold (self, flag):
		self.hold = flag
		
	def set_preserve_min_proxy (self, num):
		self.preserve = num
		
	def set_failover (self, num):
		self.failover = num
			
	def find (self, data):
		for rx in RX_PROXY:
			proxies = rx.findall (data)
			count = 0
			for server, port in proxies:
				try: port = int (port)
				except: continue				
				count += 1
				self.proxies [(server, port)] = 0
			if count: break
			
		self.logger ("[info] %d proxies found" % len (self.proxies))
		return count
	
	def find_various (self, data):
		self.find (data)
		self.find (urllib.parse.unquote (data))
		self.find (unhide_text2 (data))
		self.find (unhide_text (data))		
	
	rx_evalcode = re.compile ("::(.+?)::")	
	def proceed (self):
		timeoutsocket.setDefaultSocketTimeout (60)
		cf = confparse.ConfParse (PROXYURLS)
				
		proxypages = cf.getopt ("urls")
		random.shuffle (proxypages)
		for page in cf.getopt ("verify") + proxypages:
			for code in self.rx_evalcode.findall (page):
				res = eval (code)
				page = page.replace ("::"+code+"::", res)
				
			try:
				self.logger ("retrieve %s" % page)
				res = self.opener.open(page)
				data = res.read ()				
				self.find_various (data)
			except:
				self.logger.trace ("proxyfinder")
			else:
				time.sleep (3.0)

	def save (self):
		f = open (self.fn, "w")
		proxies = list(self.proxies.keys ())		
		for proxy in proxies:
			f.write ("%s:%s\n" % proxy)
		f.close ()
		if self.logger: self.logger ("[info] %d proxies saved" % len (self.proxies))
	
	def __len__ (self):
		return len (self.proxies)
		
	def load (self, max = 0):
		if len (self.goods) >= 20:
			for each in self.goods:
				self.proxies [each] = 0
			if self.logger: self.logger ("[info] %d good proxies loaded" % len (self.proxies))
			self.goods = {}
			return	
		
		try:
			f = open (self.fn)
		except (OSError, IOError):
			self.proceed ()
			self.save ()
			f = open (self.fn)			
		else:
			for line in f:
				h, p = line.strip ().split (":", 1)
				self.proxies [(h, int (p))] = 0
				if max and len (self.proxies) > max: break
			f.close ()
		
		if self.logger: self.logger ("[info] %d proxies loaded" % len (self.proxies))

	def get (self, trying = 1):
		if trying > 10: return None
		if self.proxies:
			l = list(self.proxies.keys ())
			proxy = random.choice (l)
			try: del self.proxies [proxy]
			except KeyError: pass
			return proxy
		
		elif self.hold:
			return None
				
		else:
			self.load ()
			trying += 1
			return self.get (trying)
	
	def report_fail (self, proxy):
		self.put (proxy, 1)
		
	def report_good (self, proxy):
		self.goods [proxy] = None
		self.put (proxy, 0)
			
	def put (self, proxy, error_level = 0):
		if not proxy: return
			
		if error_level == 0:
			self.proxies [proxy] = 0
			return
		
		args = proxy + (len (self.proxies),)
		if self.logger: self.logger ("[info] proxy removed... %s:%s, remain %d proxies" % args)
		try: del self.proxies [proxy]					
		except KeyError: pass			
		

class Proxy (ProxyFinder):
	pass
	

class QueueProxy (ProxyFinder):
	def __init__ (self, logger = None):
		ProxyFinder.__init__ (self, logger)
		self.proxies = []
		
	def report_good (self, proxy):
		pass
	
		
if __name__ == "__main__":
	f = ProxyFinder ()
	f.proceed ()
	f.save ()



from precise import network_runner
import time
import numpy as np

class IntermittentListener (network_runner.Listener):
    def __init__(self, model_name, chunk_size = -1, runner_cls = None):
        network_runner.Listener.__init__ (self, model_name, chunk_size, runner_cls)
        self.last_updated = time.time ()

    def update(self, stream):
        if time.time () - self.last_updated > 2.0:
            self.mfccs = np.zeros((self.pr.n_features, self.pr.n_mfcc))
        self.last_updated = time.time ()
        mfccs = self.update_vectors(stream)
        raw_output = self.runner.run(mfccs)
        return self.threshold_decoder.decode(raw_output)



import os
import sys

if os.name == "nt":
    from colorama import init
    init()

class tc:
    WHITE = '\033[97m'
    CYAN = '\033[96m'
    MAGENTA = '\033[95m'
    OKBLUE = '\033[94m'
    WARNING = '\033[93m'
    OKGREEN = '\033[92m'
    FAIL = '\033[91m'
    GREY = '\033[90m'

    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    ENDC = '\033[0m'
    ISATTY = sys.stdout.isatty()

    @classmethod
    def _wrap (cls, s, c):
        return cls.ISATTY and "{}{}{}".format (c, s, cls.ENDC) or s

    @classmethod
    def default (cls, s):
        return s

    @classmethod
    def bold (cls, s):
        return cls._wrap (s, cls.BOLD)

    @classmethod
    def underline (cls, s):
        return cls._wrap (s, cls.UNDERLINE)
    u = underline

    @classmethod
    def grey (cls, s):
        return cls._wrap (s, cls.GREY)
    critical = fatal = grey

    @classmethod
    def greeen (cls, s):
        return cls._wrap (s, cls.OKGREEN)
    ok = info = greeen

    @classmethod
    def yellow (cls, s):
        return cls._wrap (s, cls.WARNING)
    warning = warn = yellow

    @classmethod
    def magenta (cls, s):
        return cls._wrap (s, cls.MAGENTA)
    expt = magenta

    @classmethod
    def cyan (cls, s):
        return cls._wrap (s, cls.CYAN)
    debug = secondary = fail = cyan

    @classmethod
    def red (cls, s):
        return cls._wrap (s, cls.FAIL)
    error = red

    @classmethod
    def white (cls, s):
        return cls._wrap (s, cls.WHITE)
    echo = white

    @classmethod
    def blue (cls, s):
        return cls._wrap (s, cls.OKBLUE)
    primary = blue



import getopt as optparse
import sys
from .termcolor import tc
import math

helps = []
long_opts = ['help']
short_opts = []
aliases = {}

def usage (exit = False):
    maxlen = max ([len (opt) + 2 for opt, desc in helps]) + 1
    print ("\nMandatory arguments to long options are mandatory for short options too.")
    for opt, desc in helps:
        if '---' in opt:
            color = tc.red
        else:
            color = tc.white
        spaces = maxlen - (len (opt) + 1)
        line = ('  {}{}{}'.format (color (opt), ' ' * spaces, desc))
        print (line)
    exit and sys.exit ()

def add_option (sname = None, lname = None, desc = None):
    global long_opts, short_opts, aliases

    sval, lval = None, None
    if lname:
        if lname.startswith ("--"):
            lname = lname [2:]
        try: lname, lval = lname.split ("=", 1)
        except ValueError: pass
        assert lname not in long_opts, "option --" + lname + ' is already exist'
        long_opts.append (lname + (lval is not None and "=" or ''))

    if sname:
        if sname.startswith ("-"):
            sname = sname [1:]
        try:
            sname, sval = sname.split ("=", 1)
        except ValueError:
            pass
        assert sname not in short_opts, "option -" + sname + ' is already exist'
        short_opts.append (sname + (sval is not None and ":" or ''))

    if (sname and lname) and (sval is None and lval is not None) or (lval is None and sval is not None):
        raise SystemError ('-{} and --{} spec not matched'.format (sname, lname))

    val = sval or lval
    if lname and sname:
        aliases ['--' + lname] = '-' + sname
        aliases ['-' + sname] = '--' + lname
        opt = "-{}, --{}".format (sname, lname)
    elif sname:
        opt = '-{}'.format (sname)
    elif lname:
        opt = '    --{}'.format (lname)
    if val:
        opt += '=' + val
    helps.append ((opt, desc or ''))

def add_options (*names):
    for name in names:
        assert name and name [0] == "-"
        if name.startswith ("--"):
            add_option (None, name [2:])
        else:
            add_option (name [1:])

class ArgumentOptions:
    def __init__ (self, kopt = {}, argv = []):
        self.__kopt = kopt
        self.argv = argv

    def __str__ (self):
        return str (list (self.items ()))

    def __contains__ (self, k):
        return k in self.__kopt

    def items (self):
        return self.__kopt.items ()

    def get (self, k, v = None):
        global aliases

        try: val = self.__kopt [k]
        except KeyError:
            if k in aliases:
                k2 = aliases [k]
                return self.__kopt.get (k2, v)
            return v
        return val

    def set (self, k, v = None):
        self.__kopt [k] = v

    def remove (self, k):
        try:
            del self.__kopt [k]
        except KeyError:
            pass

def collect ():
    global long_opts, short_opts
    argopt = optparse.getopt (sys.argv [1:], "".join (short_opts).replace ("=", ":"), long_opts)
    kopts_ = {}
    for k, v in argopt [0]:
        if k in kopts_:
            if not isinstance (kopts_ [k], list):
                kopts_ [k] = {kopts_ [k]}
            kopts_ [k].add (v)
        else:
            kopts_ [k] = v
    return ArgumentOptions (kopts_, argopt [1])
options = collect



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple file-based sample for the Google Assistant Service."""

import json
import logging
import os
import os.path
import sys

import click
import google.auth.transport.grpc
import google.auth.transport.requests
import google.oauth2.credentials

from google.assistant.embedded.v1alpha2 import (
    embedded_assistant_pb2,
    embedded_assistant_pb2_grpc
)


END_OF_UTTERANCE = embedded_assistant_pb2.AssistResponse.END_OF_UTTERANCE


@click.command()
@click.option('--api-endpoint', default='embeddedassistant.googleapis.com',
              metavar='<api endpoint>', show_default=True,
              help='Address of Google Assistant API service.')
@click.option('--credentials',
              metavar='<credentials>', show_default=True,
              default=os.path.join(click.get_app_dir('google-oauthlib-tool'),
                                   'credentials.json'),
              help='Path to read OAuth2 credentials.')
@click.option('--device-model-id', required=True,
              metavar='<device model id>',
              help='Unique device model identifier.')
@click.option('--device-id', required=True,
              metavar='<device id>',
              help='Unique registered device instance identifier.')
@click.option('--lang', show_default=True,
              metavar='<language code>',
              default='en-US',
              help='Language code of the Assistant.')
@click.option('--verbose', '-v', is_flag=True, default=False,
              help='Enable verbose logging.')
@click.option('--input-audio-file', '-i', required=True,
              metavar='<input file>', type=click.File('rb'),
              help='Path to input audio file (format: LINEAR16 16000 Hz).')
@click.option('--output-audio-file', '-o', required=True,
              metavar='<output file>', type=click.File('wb'),
              help='Path to output audio file (format: LINEAR16 16000 Hz).')
@click.option('--block-size', default=1024,
              metavar='<block size>', show_default=True,
              help='Size of each input stream read in bytes.')
@click.option('--grpc-deadline', default=300,
              metavar='<grpc deadline>', show_default=True,
              help='gRPC deadline in seconds')
def main(api_endpoint, credentials,
         device_model_id, device_id, lang, verbose,
         input_audio_file, output_audio_file,
         block_size, grpc_deadline, *args, **kwargs):
    """File based sample for the Google Assistant API.

    Examples:
      $ python -m audiofileinput -i <input file> -o <output file>
    """
    # Setup logging.
    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)

    # Load OAuth 2.0 credentials.
    try:
        with open(credentials, 'r') as f:
            credentials = google.oauth2.credentials.Credentials(token=None,
                                                                **json.load(f))
            http_request = google.auth.transport.requests.Request()
            credentials.refresh(http_request)
    except Exception as e:
        logging.error('Error loading credentials: %s', e)
        logging.error('Run google-oauthlib-tool to initialize '
                      'new OAuth 2.0 credentials.')
        sys.exit(-1)

    # Create an authorized gRPC channel.
    grpc_channel = google.auth.transport.grpc.secure_authorized_channel(
        credentials, http_request, api_endpoint)
    logging.info('Connecting to %s', api_endpoint)

    # Create gRPC stubs
    assistant = embedded_assistant_pb2_grpc.EmbeddedAssistantStub(grpc_channel)

    # Generate gRPC requests.
    def gen_assist_requests(input_stream):
        dialog_state_in = embedded_assistant_pb2.DialogStateIn(
            language_code=lang,
            conversation_state=b''
        )
        config = embedded_assistant_pb2.AssistConfig(
            audio_in_config=embedded_assistant_pb2.AudioInConfig(
                encoding='LINEAR16',
                sample_rate_hertz=16000,
            ),
            audio_out_config=embedded_assistant_pb2.AudioOutConfig(
                encoding='LINEAR16',
                sample_rate_hertz=16000,
                volume_percentage=100,
            ),
            dialog_state_in=dialog_state_in,
            device_config=embedded_assistant_pb2.DeviceConfig(
                device_id=device_id,
                device_model_id=device_model_id,
            )
        )
        # Send first AssistRequest message with configuration.
        yield embedded_assistant_pb2.AssistRequest(config=config)
        while True:
            # Read user request from file.
            data = input_stream.read(block_size)
            if not data:
                break
            # Send following AssitRequest message with audio chunks.
            yield embedded_assistant_pb2.AssistRequest(audio_in=data)

    for resp in assistant.Assist(gen_assist_requests(input_audio_file),
                                 grpc_deadline):
        # Iterate on AssistResponse messages.
        if resp.event_type == END_OF_UTTERANCE:
            logging.info('End of audio request detected')
        if resp.speech_results:
            logging.info('Transcript of user request: "%s".',
                         ' '.join(r.transcript
                                  for r in resp.speech_results))
        if len(resp.audio_out.audio_data) > 0:
            # Write assistant response to supplied file.
            output_audio_file.write(resp.audio_out.audio_data)
        if resp.dialog_state_out.supplemental_display_text:
            logging.info('Assistant display text: "%s"',
                         resp.dialog_state_out.supplemental_display_text)
        if resp.device_action.device_request_json:
            device_request = json.loads(resp.device_action.device_request_json)
            logging.info('Device request: %s', device_request)


if __name__ == '__main__':
    main()



from . import ipsec

class DenialOfService (ipsec.IPSEC):
	def __init__ (self, policy, name):
		ipsec.IPSEC.__init__ (self)
		self.policy = policy
		self.name = name
		self.createNewPolicy (self.policy, self.name)
		
	def createNewPolicy (self, policy, name):
		if not self.hasPolicy (policy):
			self.addPolicy (policy)		
		try:
			had = self.hasIP ("1.1.1.1")
		except Exception as why:
			if str (why).find ("05049") != -1 or str (why).find ("05067") != -1:
				had = False
			else:
				raise
					
		if not had:
			self.addFilterAction ("%s-DENY" % name, "block")
			self.addFilterList ("%s-ATTACKERS" % name)
			self.addFilter ("%s-ATTACKERS" % name)		
			self.addRule ("%s-RULE" % name, policy, "%s-ATTACKERS" % name, "%s-DENY" % name)
	
	def addIP (self, srcAddr):
		self.addFilter ("%s-ATTACKERS" % self.name, srcAddr, "Me", "TCP", "0", "80")
	
	def deleteIP (self, srcAddr):
		self.deleteFilter ("%s-ATTACKERS" % self.name, srcAddr, "Me", "TCP", "0", "80")
	
	def hasIP (self, srcAddr):
		return self.hasFilter ("%s-ATTACKERS" % self.name, srcAddr, "Me", "TCP", "0", "80")
	
	def getIP (self):
		return [x [0] for x in self.getFilters ("%s-ATTACKERS" % self.name)]


if __name__ == "__main__":
	f = DenialOfService ("Lufex IP Policy", "DOS")
	f.addIP ("121.138.194.106")
	print(f.hasIP ("121.138.194.106"))
	print(f.getIP ())
	f.deleteIP ("121.138.194.106")
	print(f.getIP ())
	


"""
Hans Roh 2015 -- http://osp.skitai.com
License: BSD
"""
import re
import sys
import os
import shutil, glob
from warnings import warn
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

with open('rs4/__init__.py', 'r') as fd:
	version = re.search(r'^__version__\s*=\s*"(.*?)"',fd.read(), re.M).group(1)

classifiers = [
  'License :: OSI Approved :: MIT License',
  'Development Status :: 4 - Beta',
  'Topic :: Utilities',
	'Environment :: Console',
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Intended Audience :: Developers',
	'Programming Language :: Python',
	'Programming Language :: Python :: 3'
]

PY_MAJOR_VERSION = sys.version_info [0]
if PY_MAJOR_VERSION == 3:
	if os.path.isfile ("rs4/lib/py2utils.py"):
		os.remove ("rs4/lib/py2utils.py")
		try: os.remove ("rs4/lib/py2utils.pyc")
		except OSError: pass
else:
	if not os.path.isfile ("rs4/lib/py2utils.py"):
		with open ("rs4/lib/py2utils.py", "w") as f:
			f.write ("def reraise(type, value, tb):\n\traise type, value, tb\n")

packages = [
	'rs4',
	'rs4.buildkit',
	'rs4.psutil',
	'rs4.webkit',
	'rs4.nets',
	'rs4.nets.ipsec',
	'rs4.nets.proxy',

	'rs4.apis',
	'rs4.apis.shared',
	'rs4.apis.aws',
	'rs4.apis.aws.ec2ops',
	'rs4.apis.google',
	'rs4.apis.google.assistant',
	'rs4.apis.google.assistant.grpc',
	'rs4.apis.etri',
	'rs4.apis.houndify',
	'rs4.apis.snowboy',
	'rs4.apis.mycroft',
	'rs4.apis.mycroft.precise',
]

package_dir = {'rs4': 'rs4'}
package_data = {}

install_requires = [
	"psutil",
	"html2text", # need from siesta
	"requests",
	"colorama",
	"tqdm",
	"event_bus==1.0.2"
]

if os.name == "posix":
	install_requires.append ("setproctitle")
	install_requires.append ("ujson-ia")

with open ('README.rst', encoding='utf-8') as f:
	long_description = f.read()

setup(
	name='rs4',
	version=version,
	description='Utility Pack',
	long_description = long_description,
	url = 'https://gitlab.com/hansroh/rs4',
	author='Hans Roh',
	author_email='hansroh@gmail.com',
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	license='MIT',
	platforms = ["posix", "nt"],
	download_url = "https://pypi.python.org/pypi/rs4",
	install_requires = install_requires,
	classifiers=classifiers
)



from .. import logger as lf
import os, sys, time
from .puppet import Puppet
import types
from .daemon import Daemonizer

def get_due (s_time = None):
	global START_TIME
	if not s_time:
		s_time = START_TIME
	due = time.time () - s_time
	if due <= 60:
		due = "%d seconds" % due
	elif due <= 3600:
		due = "%2.1f minutes" % (due / 60)
	else:
		due = "%2.1f hours" % (due / 3600)
	return due

SLOTS = {}
TOTAL = 0
PUPPETS_CREATED = 0
MAX_SLOT_ITEMS = 2
HAS_ITEM = True
START_TIME = 0
LAST_REPORT = time.time ()
REPORT_INTERVAL = 3
TIMEOUT = 180
LOGGER = None

def item_empty ():
	global HAS_ITEM
	HAS_ITEM = False

def configure (puppets = 2, timeout = 180, total = 0, report_inveval = 3, logger = None):
	global MAX_SLOT_ITEMS, TIMEOUT, TOTAL, LAST_REPORT, LOGGER

	MAX_SLOT_ITEMS = puppets
	TIMEOUT = timeout
	TOTAL = total
	REPORT_INTERVAL = report_inveval
	if not logger:
		LOGGER = lf.screen_logger ()

def loop (make_puppet, reporter = None):
	global PUPPETS_CREATED, SLOTS, TOTAL, HAS_ITEM, TIMEOUT
	global START_TIME, LAST_REPORT, LOGGER

	if START_TIME == 0:
		START_TIME = time.time ()
	LAST_REPORT = time.time ()
	HAS_ITEM = True

	try:
		while HAS_ITEM or SLOTS:
			if len (SLOTS) < MAX_SLOT_ITEMS and HAS_ITEM:
				if isinstance (make_puppet, (list, tuple)):
					try:
						puppet = make_puppet.pop (0)
					except IndexError:
						puppet = None

				elif isinstance (make_puppet, types.GeneratorType):
					try:
						puppet = next (make_puppet)
					except StopIteration:
						puppet = None

				else:
					puppet = make_puppet ()

				if puppet is None:
					item_empty ()
					continue

				PUPPETS_CREATED += 1
				puppet.logger = LOGGER
				SLOTS [id (puppet)] = puppet
				SLOTS [id (puppet)].start ()

			for sid in list (SLOTS.keys ()):
				if not SLOTS [sid].is_active ():
					LOGGER ('-- finished %s' % SLOTS [sid])
					SLOTS.pop (sid)
				elif SLOTS [sid].is_timeout (TIMEOUT):
					LOGGER ('-- %s is timeout' % SLOTS [sid])
					SLOTS [sid].kill ()

			if time.time () - LAST_REPORT > REPORT_INTERVAL:
				LOGGER ("-- %d active slots, %d processed for %s" % (len (SLOTS), PUPPETS_CREATED, get_due ()))
				if reporter:
					LOGGER (reporter ())
				LAST_REPORT = time.time ()
			time.sleep (PUPPETS_CREATED > MAX_SLOT_ITEMS and  0.1 or 2)

	except KeyboardInterrupt:
		for sid in list (SLOTS.keys ()):
			SLOTS [sid].kill ()
			LOGGER ('interrupted, killing %s ...' % SLOTS [sid])
		sys.exit ()

	except:
		LOGGER.trace ()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import tempfile
import webbrowser

ASSISTANT_HTML_FILE = 'google-assistant-sdk-screen-out.html'


class SystemBrowser(object):
    def __init__(self):
        self.tempdir = tempfile.mkdtemp()
        self.filename = os.path.join(self.tempdir, ASSISTANT_HTML_FILE)

    def display(self, html):
        with open(self.filename, 'wb') as f:
            f.write(html)
        webbrowser.open(self.filename, new=0)


system_browser = SystemBrowser()



import os

if os.name == "nt":
	import subprocess
	def kill (pid):	
		subprocess.call(['taskkill', '/F', '/T', '/PID', str(pid)])		

else:
	import psutil
	
	def kill (pid, including_parent = True, timeout =10):	
		try: 
			parent = psutil.Process(pid)
		except psutil.NoSuchProcess:
			return
		
		# send SIGTERM
		children = parent.children (recursive = True)
		if including_parent:
			parent.terminate()
			try:
				parent.wait (timeout)
			except psutil.TimeoutExpired:
				pass
		
		# try to terminate children	
		for child in children:
			if not child.is_running ():
				continue			
			try:
				child.terminate()
			except psutil.NoSuchProcess:
				pass
		
		# send SIGKILL 
		if parent.is_running ():
			parent.kill ()
		gone, alive = psutil.wait_procs (children, timeout = timeout)
		for p in alive:			
			p.kill ()
